# encoding: utf-8

from yade import ymport, pack, plot, export  # export-paraview, ymport-sampleGen, qt-graphics, pack-sample packing
from yade.gridpfacet import *  # * -> imports the content into the normal namespace
import sys  # -> .stdout.flush() - use to flush print output to batch log files
import math  # for math.isnan()
import os  # for simulation folder creation
import pyvtk  # for creating vtk legacy file series

#import builtins  # <- for python 3. Equiv for python 2 -> __builtin__
import __builtin__
import datetime # for date and time labeling
import time
start_time = time.time()

# Give tabledata as input when running batch mode for parametric studies
nRead=readParamsFromTable(  # Read values from table. If no table given-> use the following parameters:
    finalFricDegree = 30,
    densityScaleFactor = 1e0,
    young = 200e6,
    wallVelocity = 0.1,
    damping=0.03,
)
from yade.params import table   # import from table -> use table."parametername" to access values
#from yade.params.table import *  # import from table into current namespace

# Add table parameters
densityScaleFactor=table.densityScaleFactor

## for importing txt file parameters generated from sampleGen
import imp

importedParam=0
paramFile=open("1paramSampleGen.txt")
param = imp.load_source("1paramSampleGen.txt","",paramFile)
if not paramFile.closed:
    importedParam=1
paramFile.close()

###################################
#####  Simulation settings  #####
###################################

view3D = 0
plotData = 1  # addPlotData and save to file
showPlot = 1  # draw plots in realtime

captureData = 0 # save plot data and input parameters in txt files
captureParaviewData = 1  # capture forceChains, rotations, displacements etc
captureLocalPorosityData = 0  # save grid mesh local porosity from center of sample
captureIterPeriod = 200  # periodic capture of data for plots and print of strain every X number of iteration

# specify at what vertical strain levels to capture displacement data
strainIncrements = [0.0025*i for i in range(1,1+4*20)]

calcPorosity = 1  # determine porosity of representative inner volume of sample to avoid boundary effects near membrane
tesselatedMicroStrain = 0  # tesselate sample and calculate strain between strainIncrements given above

batchMode=0
if runningInBatch():
    captureParaviewData=0
    batchMode = 1  ### Not added fully

####  Membrane settings  ####
numMembranes = 4  # 2 gives just the membranes parallel to y-axis for genMembranes == 1
genMembrane = 3
# 1-separate membrane for each boundary     2-continuous membrane     3 - Shifted triangular grid -> more geom. flexible
genPFacets = 0  # 1 - filled in PFacets membranes        # 0 - rectangular grid membranes
rigidWalls = 1  # 1 -> generate 2 rigid walls, and 2 membranes          2 -> use 4 rigid walls
planeStrainConditions= 0  # 1- disable stress control of walls in y-dir

verticalLock=0  # lock membrane nodes dof in the z-direction
parallelLock=0  # lock membrane nodes dof parallel to membra

genAnchors = 0
# 1 - for 4 separate membranes, add fixed nodes above and below
anchorFacets = 0
# 1 - extend membrane up "extendedMembraneRows" rows ###   ? lock dofs?
lockAnchors = 0  # lock anchornode dofs in xyz
lockAnchorsZ = 0   # lock anchornode dofs in z
lockTopAnchorsZ = 0  # lock z dof of the top edge of the membrane
lockTopAnchorsX = 0  # lock x dof of the top edge of the membrane
extendedMembraneRows = 3  # for genMembrane==1

# facets to block the membranes from deforming when not in contact with sample
genCovFacets = 2  # 1 - covers of P-facets      2 - covers of gridConnections
genSideCovFacets = 1
genTopBotCovFacets = 1
genFullBlock = 0  # pfacet cover to block the whole membrane -> to check force application
moveCoversWithYWalls=1

coverFriction = 0  # 1 - friction between side covers and membrane
membraneFriction=0  # 1 - full roughness on membrane

topCoverFollowStrain=1  # move topCover (and botCover) with the vertically straining loadPlates
strainBothWays=1  # Strain also from bot -> set bottom plate vel = top plate vel = wallVelocity/2
topBotFriction = 1 # Full roughness on top and bot load plates

engiStress=0
# 1 - stress based on orig area of sample in xy plane      0 - account for strain of rigid walls in stress calc

# set body velocity to zero at intervals
calmFunction = 0
# 1 - calm() continuously    2 - calm() in initial iterations when membrane comes into contact
calmIterPeriod = 10  # iterPeriod for continuous calm function
calmSteps = 1000  # initial steps with calm()

####  Strain and lateral stress settings  ####
compress = 1
# 1 - add compression phase

activeNodesPadding = 0  # apply force on nodes adjacent to active area as well
lateralStress = 3
# 1 - turn on lateral stress       2 - mass adjusted force on all nodes(incl. anchors) -> use with moving top covers
# 3 - force on nodes inside area confined by the 4 rigid walls in z and y dir
massDistributedLoad=1  # lump node force dep on the node mass
pressureNormToMem=0  # apply membrane force normal to each triangle area (for lateralStress==3)

borderNodesHalfLoad = 0
# 1 - load borders nodes equally 1/2       # 0 - load nodes 1/3 and 2/3

clumpSample = 0
genSpheres = 1
# 1 - Load sample from "sampleGen.py" ,     2 - makeCloud deposited into box from height,     3 - a few particles for timestep
# 4 - Loose cubical sample deposit     5 - loose cubical sample
# 6 - half width sample in center to stop shrinking membrane
# 7 - randomDensePack - Not working properly

###################################
##### User defined parameters #####
###################################

disableGlobalStiffTimeStep=1  # if True - disable and specified time step instead
timeStep =6e-7  # 8e-7
adjustTimestepSafety = 0.18 #0.18 for shallow 13k # 0.04 for 8k
                            # 0.05 for 8k sphere
if rigidWalls==2:
    adjustTimestepSafety=0.8

# with membrStiff=8e6 -> stable at 0.02
timestepSafety = 0.0000001   # initial small time step for stability - perform 1 initial step for lateral stress calc!

sigmaIso = -200e3 # Pa
targetStrain = -.2  # vertical strain target
gravity = (0,0,0)  # z: -9.81

# for pFacets -> interactions with spheres
youngFacet=3e10     # originalt 1e7
densityFacet=100*densityScaleFactor  # originalt 2650
densityGridNode=200*densityScaleFactor

# for the membrane
momRotLaw=0  # turns on momentresistance in pfacet/connection nodes proportional to young
# if momRotLaw == 0 -> nodes in a series of connections act a links

membraneInc =1.0 #.6 - membrane width larger than sample -> use float number!

# number of voxels for porosity calculation
poroResolution = 400

if not importedParam:
    widthX = width; widthY = width
    wallStartPos=[0,width,0,width,0,height]
    sigmaX = sigmaY = sigmaZ = sigmaIso


if importedParam:
    # sample parameters imported from sampleGen.py
    wallStartPos=param.wallFinalPos
    clumpSample=param.clumpSample
    shallowSample=param.shallowSample
    youngWalls=param.youngWalls
    numParticles=param.numParticles
    height=param.height
    widthX=param.widthX
    widthY=param.widthY
    sigmaIso=param.sigmaIso
    sigmaX=param.sigmaX
    sigmaY=param.sigmaY
    sigmaZ=param.sigmaZ
    # particle parameters
    young = param.young
    poisson=param.poisson
    density=param.density
    compFricDegree=param.compFricDegree
    rMean=param.rMean
    rRelFuzz=param.rRelFuzz
    compactedPorosity=param.compactedPorosity

    # EVT.:
    widthX = wallStartPos[1]-wallStartPos[0]
    widthY = wallStartPos[3]-wallStartPos[2]
    height = wallStartPos[5]-wallStartPos[4]


# Add table parameters
densityScaleFactor=table.densityScaleFactor
wallVelocity = table.wallVelocity
finalFricDegree = table.finalFricDegree
damping=table.damping
if batchMode:
    young = table.young


width=widthX

# adjust membrane:
if shallowSample:
    # For shallow -0.3 -> nw = 10; nh = 15
    nw = 10
    nh = 15
    membraneStiff = 9e4
    gridNormStiff = 250
    if widthY == 0.02:
        nw = 8
        nh = 16
        membraneStiff = 10e4
        if height==0.25:
            nh = 19

else:
    if height==2*widthX:
        # For normal dim -> nw = 13; nh = 9
        nw = 13  # 13  # num nodes along perimeter
        nh = 9  # 9   # num nodes along height
        membraneStiff = 4.7e4
        gridNormStiff = 250
    if height==widthX:
        nw=12
        nh=4
        membraneStiff = 6e4


# scale density for stability
density*=densityScaleFactor

if shallowSample:
    membraneInc *= 2

widthM1 = (1+membraneInc)*widthY  # width of membrane normal to to x-axis
widthM2 = (1+membraneInc)*widthX  # width of membrane normal to to y-axis
#widthM = (1+membraneInc)*width

membrVertExt = 0.1 * height # 0.2 * # extend membrane above and below sample by this much for genMembrane==3

radiusGN=width/40 #  radius of grid node/ wall thickness -> larger = more stable (due to larger mass/inertia?)

## Correct some dependent settings
# if runningInBatch():
#     batchMode=1
if not batchMode and view3D:  #disable graphics for server and batch
    from yade import qt
if batchMode:
    view3D=0;  showPlot=1;  captureData=0;  plotData=1
if genMembrane==2:
    rigidWalls=0; numMembranes=4
if rigidWalls:
    numMembranes=2
    #nw=2
if nw==2:
    genSideCovFacets=0
if rigidWalls==2:
    genMembrane=0; lateralStress=0; genAnchors=0; genCovFacets=0
if genMembrane==3:
    rigidWalls=1; genAnchors=0
    if lateralStress == 1: print("Incorrect stress mode! Use lateralStress = 2 or 3")
if not genCovFacets:
    topCoverFollowStrain=0
    genSideCovFacets=0
    genTopBotCovFacets=0

if not genPFacets:
    if genMembrane == 3:
        heightM = height + membrVertExt * 2
        pfacetH = heightM / (nh - 1)
        pfacetW = widthM1 / (nw - 1)
        if pfacetH > pfacetW:
            radiusGN = pfacetW / 2 * 0.8  #0.68
        else:
            radiusGN = pfacetH / 2 * 0.8

    else:
        pfacetH=height/(nh-1)
        pfacetW=widthM1/(nw-1)
        if pfacetH > pfacetW:
            radiusGN=pfacetW/2*0.85
        else:
            radiusGN=pfacetH/2*0.85

# reach iter target to maintain stability before starting deviatoric phase
iterTargetFlex =  200# 1000 + 0.6 * numParticles #(12000 + 0.5 * numParticles)  * sqrt(young*densityScaleFactor/(300e6)) * 0.2/adjustTimestepSafety  # stiffness and density scaling change timestep
iterTargetRigid = 4000 * sqrt(young*densityScaleFactor/(300e6))

# Adjust data capture period depending on densityScaling,particle stiffness and strain rate
if rigidWalls==1:
    iterAdjustFactor = sqrt(young/(300e6*densityScaleFactor)) * 0.8/wallVelocity
elif rigidWalls==2:
    iterAdjustFactor = sqrt(young / (300e6 * densityScaleFactor)) * 0.8 / wallVelocity * 0.1

captureIterPeriod=int(captureIterPeriod * iterAdjustFactor)

###################################
##### Materials #####
###################################
# Particle material
O.materials.append(FrictMat(young=young, poisson=poisson, frictionAngle=radians(finalFricDegree),
                            density=density*densityScaleFactor, label='spheres'))

## Same material for facet walls as in sampleGen.py
# Material for frictionless support walls
O.materials.append(FrictMat(young=youngWalls,poisson=poisson,frictionAngle=radians(0), density=0,
                            label='walls'))
# Material for top and bot load plates with full friction
wallZMatId = O.materials.append(FrictMat(young=youngWalls,poisson=poisson, frictionAngle=radians(finalFricDegree),
                                         density=0, label='zWalls'))

## Membrane materials
membFricDegree=0
if membraneFriction:
    membFricDegree=finalFricDegree

O.materials.append(CohFrictMat(young=membraneStiff,poisson=.3,density=densityGridNode,frictionAngle=radians(0),
                               normalCohesion=3e15, shearCohesion=3e15,momentRotationLaw=momRotLaw,label='gridNodeMat'))

O.materials.append( FrictMat(young=youngFacet,poisson=.3,density=densityFacet,frictionAngle=radians(membFricDegree),
                             label='pFacetMat'))

######################################
#####   Create confining facets #####
######################################

# define corners of the sample packing
mn,mx=Vector3(wallStartPos[0],wallStartPos[2],wallStartPos[4]), Vector3(wallStartPos[1],wallStartPos[3],wallStartPos[5])

# determine oversizeFactor for facets
oversizeFactor=sigmaY/sigmaX*10
if sigmaX==sigmaY==sigmaZ:  # isotropic stress -> smaller load plates
    oversizeFactor=2
## create walls around the packing
walls=aabbWalls([mn,mx],thickness=0,oversizeFactor=oversizeFactor,material='walls')
O.bodies.append(walls)  ##Prior: wallIds=O.bodies.append(walls)

if topBotFriction:
    # assign friction on top and bot walls
    #top
    walls[5].mat=O.materials[wallZMatId]
    #bot
    walls[4].mat=O.materials[wallZMatId]

## ADD FACET IDS
wallXMin=walls[0]
wallXMax=walls[1]
wallYMin=walls[2]
wallYMax=walls[3]
bottomPlate=walls[4]
topPlate=walls[5]

############

wallBodies=[wallXMin,wallXMax,wallYMin,wallYMax,bottomPlate,topPlate]
if rigidWalls==1 or rigidWalls==0:
    O.bodies.erase(wallXMax.id)
    O.bodies.erase(wallXMin.id)
    if rigidWalls == 0:
        O.bodies.erase(wallYMax.id)
        O.bodies.erase(wallYMin.id)


#########################################
#####   Generate sample   #####
#########################################
clumpIds=[]             # list of clump ids
clumpMemberIds=[]       # list of clump member ids

# load particles
if genSpheres==1:
    if clumpSample:
        sp = ymport.textClumps('compactSample.txt', material="spheres")  # loads and inserts the clumps

        # list containing all ids of loaded clump particles + individual clump member spheres
        sphereIds = [id for id in range(6, len(O.bodies))]  # the first 6 spaces are taken by
        # walls (also the erased ones)
        O.bodies.updateClumpProperties()  # update to get correct volume, mass, inertia of clumps

        ## sort out clumps and clumpMembers
        for i in sphereIds:
            if O.bodies[i].isClump:
                clumpIds.append(i)
            else:
                clumpMemberIds.append(i)

    else:
        sp = ymport.text('compactSample.txt',material="spheres")  # load spheres
        for s in sp:    # insert into simulation
            O.bodies.append(s)

        # list containing all ids of loaded spheres
        sphereIds = [id for id in range(6,len(O.bodies))]

    # # number of loaded particles OR clumps+clumpMembers
    # numParticles=len(sphereIds)

    for s in O.bodies:
        # reset ref pos of all spheres to current pos in order to display displacement later -> O.bodies[id].state.displ()
        s.state.refPos = s.state.pos
        # reset ref rotation to display rotation later -> O.bodies[id].state.rot()
        s.state.refOri = s.state.ori


############### For initial testing without generated sample
if genSpheres==2: # makecloud deposit into membrane
    # create empty sphere packing
    # sphere packing is not equivalent to particles in simulation, it contains only the pure geometry
    sp=pack.SpherePack()
    # generate randomly spheres with uniform radius distribution
    sp.makeCloud((-.03,-.03,0),(.03,.03,.45),rMean=.006,rRelFuzz=.5)
    sp.toSimulation(material='spheres')

elif genSpheres == 3:   # small amount of particles for timestep stability
    sampleToWidth=0.8
    sp=pack.SpherePack()
    # generate randomly spheres with uniform radius distribution
    sp.makeCloud(((1-sampleToWidth)*width,(1-sampleToWidth)*width,0),(width*sampleToWidth,width*sampleToWidth, height),
                 rMean=.006,rRelFuzz=.5, seed=1)
    sp.toSimulation(material='spheres')

elif genSpheres == 4:   # loose cubic sample deposit 2 x sample height
    sampleToWidth=0.98
    sp=pack.SpherePack()
    # generate randomly spheres with uniform radius distribution
    sp.makeCloud(((1-sampleToWidth)*width,(1-sampleToWidth)*width,0),(width*sampleToWidth,width*sampleToWidth, 2*height),
                 rMean=.006,rRelFuzz=.5, seed=1)
    sp.toSimulation(material='spheres')

elif genSpheres == 5:   # loose cubic sample with correct dimensions
    sampleToWidth=0.98
    sp=pack.SpherePack()
    # generate randomly spheres with uniform radius distribution
    sp.makeCloud(((1-sampleToWidth)*width,(1-sampleToWidth)*width,0),(width*sampleToWidth,width*sampleToWidth, height),
                 rMean=.006,rRelFuzz=.5, seed=1)
    sp.toSimulation(material='spheres')

elif genSpheres==6:     # half width sample to observe membrane shrinking and interaction
    sampleToWidth=0.8
    sp=pack.SpherePack()
    # generate randomly spheres with uniform radius distribution
    sp.makeCloud(((1-sampleToWidth)*width,(1-sampleToWidth)*width,height*0.05),(width*sampleToWidth,width*sampleToWidth, height*0.95),
                 rMean=.006,rRelFuzz=.5, seed=1)
    sp.toSimulation(material='spheres')

elif genSpheres==7:     # random dense pack - not working!
    # Not working properly!
    pred=pack.inAlignedBox((0,0,0),(.1,.1,.1))
    spheres=pack.randomDensePack(pred, radius=0.006, cropLayers=0, rRelFuzz=0, spheresInCell=200, memoizeDb=None,
                                 useOBB=0, memoDbg=0, color=None, returnSpherePack=True)

#####################################
#####  Setup for data capture  #####
#####################################
path = ""
if captureData:
    dateTime = datetime.datetime.now()
    date = "_"+str(dateTime.day)+"-"+str(dateTime.month)+"_"
    clock = str(dateTime.hour)+":"+str(dateTime.minute)
    ## create folder for data
    # name of data folder
    folderName = "simData"+date+clock ## + time and date
    path = folderName
    try:
        os.mkdir(path)
    except OSError:
        print ("Creation of the directory %s failed" % path)
    else:
        print ("Successfully created the directory %s " % path)

    path+="/"  # to simplify file naming

    # data capture engine for Paraview - saved to temporary files!
    vtkPart = export.VTKExporter(path+'sampleData')        #  particle data
    vtkMem = export.VTKExporter(path+'membraneNodeData')   #  membrane data
    vtkShear = export.VTKExporter(path+'sampleDataShear')
    #vtkDispl = export.VTKExporter(path+'displData')


###########################
##### ENGINES #####
###########################
O.engines=[
    ForceResetter(),
    InsertionSortCollider([
        Bo1_Sphere_Aabb(),
        Bo1_Wall_Aabb(),
        Bo1_PFacet_Aabb(),
        Bo1_Node_Aabb(),
        Bo1_GridConnection_Aabb(),
        Bo1_Box_Aabb(),             # for imported facetBox from sampleGen instead of walls
    ]),
    #sortThenCollide=1),
    InteractionLoop([
        Ig2_PFacet_PFacet_ScGeom(),     # for interaction between membranes and covers
        Ig2_GridNode_GridNode_GridNodeGeom6D(),
        Ig2_GridConnection_GridConnection_GridCoGridCoGeom(),
        Ig2_Sphere_PFacet_ScGridCoGeom(),           # pFacets
        #Ig2_Sphere_Sphere_ScGeom6D(),
        Ig2_Sphere_Sphere_ScGeom(),
        #Ig2_Facet_Sphere_ScGeom(),       # Facet
        Ig2_Sphere_GridConnection_ScGridCoGeom(),
        Ig2_Box_Sphere_ScGeom(),        # for imported facetBox from sampleGen instead of walls

        Ig2_Wall_Sphere_ScGeom(),        # interaction with bottom plate and load plate
        #Ig2_Wall_PFacet_ScGeom(),       # interaction between plate and membrane

        Ig2_GridConnection_PFacet_ScGeom(), # fix weird membrane - bottom plate interaction
    ],
    [
        Ip2_CohFrictMat_CohFrictMat_CohFrictPhys(setCohesionNow=1,setCohesionOnNewContacts=0),
        Ip2_FrictMat_FrictMat_FrictPhys()   # physics for external interactions, i.e., cylinder-cylinder,
                                            # sphere-sphere, cylinder-sphere
    ],
    [
        Law2_ScGeom6D_CohFrictPhys_CohesionMoment(),        # contact law for "internal" cylinder forces
        Law2_ScGeom_FrictPhys_CundallStrack(),              # contact law for sphere-sphere
        Law2_ScGridCoGeom_FrictPhys_CundallStrack(),        # contact law for cylinder-sphere
        Law2_GridCoGridCoGeom_FrictPhys_CundallStrack()     # contact law for cylinder-cylinder interaction
    ]
    ),
    GlobalStiffnessTimeStepper(active=1,timeStepUpdateInterval=10,
                               timestepSafetyCoefficient=timestepSafety,label='ts'),
    PyRunner(command='confiningPres()', iterPeriod=1, label='confiningEngine'),
    PyRunner(command='rigidBoundaryController()', iterPeriod=1, label='rigidController'),
    NewtonIntegrator(gravity=gravity,damping=damping,label='newton'),

    PyRunner(command='checkUnbalanced()',iterPeriod=captureIterPeriod/6,label='checker'),  # Check if stable -> apply vertical strain -> monitor vertical strain against strain goals

    # save images from Yade's 3d view
	#qt.SnapshotEngine(fileBase='/test/3d-',iterPeriod=1000,label='snapshot'),

    # PRINT INFO:
    PyRunner(iterPeriod=captureIterPeriod*40, label='infoPrinter',command="printParam()"),

    # add data to plot
    PyRunner(iterPeriod=captureIterPeriod, label='plotter',command="addPlotData()"),
]


### Disable selected engines ###
if disableGlobalStiffTimeStep:
    ts.dead=1
    if rigidWalls==1:
        O.dt=timeStep
    elif rigidWalls==2:
        O.dt=0.8*PWaveTimeStep()

if not plotData:
    plotter.dead=1

if lateralStress == 0:
    confiningEngine.dead = 1

if lateralStress == 2:
    confiningEngine.command="confiningPresAll()"

if lateralStress == 3:
    pass


######################################
#####  Data Capture Functions  #####
######################################

#### Interaction Tracking ####
def forceChains():
    # Preallocate
    global selectInteractions
    global interactionIds
    selectInteractions = []
    interactionIds = []
    #### Single out particle interactions:
    for i in O.interactions:
        # # skip gridNode interactions
        if isinstance(O.bodies[i.id1].shape,GridNode) or isinstance(O.bodies[i.id2].shape,GridNode):
            continue
        # skip pFacet interactions
        if isinstance(O.bodies[i.id1].shape, PFacet) or isinstance(O.bodies[i.id2].shape, PFacet):
            continue
        # skip GridConnections
        if isinstance(O.bodies[i.id1].shape, GridConnection) or isinstance(O.bodies[i.id2].shape, GridConnection):
            continue
        # skip wall interactions
        elif isinstance(O.bodies[i.id1].shape,Wall) or isinstance(O.bodies[i.id2].shape,Wall):
            continue
        # skip facetBox interactions
        elif isinstance(O.bodies[i.id1].shape,Box) or isinstance(O.bodies[i.id2].shape,Box):
            continue

        # Skip clump internal connection interactions
        if O.bodies[i.id1].isStandalone==False and O.bodies[i.id1].clumpId==O.bodies[i.id2].clumpId:
            continue

        else:
            selectInteractions.append(i)
    interactionIds = [(i.id1,i.id2) for i in selectInteractions]
    # capture data
    vtkPart.exportInteractions(ids=interactionIds, what=[('forceN','i.phys.normalForce.norm()')])
    #vtkShear.exportInteractions(ids=interactionIds, what=[('forceS','i.phys.shearForce.norm()')])

    if genMembrane!=0:
        ### capture membrane gridNode data -> visualize membrane by itself
        selectInteractions=[]  # reset
        global membraneInteractions
        membraneInteractions = []
        for i in O.interactions:
            # add gridNode interactions. No need to check id.2 as gridNodes only interact with other gridNodes
            if isinstance(O.bodies[i.id1].shape, GridNode) and (O.bodies[i.id1].mask==5 or O.bodies[i.id1].mask==2):
                selectInteractions.append(i)
        membraneInteractions = [(i.id1, i.id2) for i in selectInteractions]
        vtkMem.exportInteractions(ids=membraneInteractions, what=[('forceN','i.phys.normalForce.norm()')])



####  Rotation Tracking  ####
def rotationTracker():
    if clumpSample:
        vtkPart.exportSpheres(ids=clumpMemberIds, useRef=0,
                          what=[('rot', 'b.state.rot().norm()')])
    else:
        vtkPart.exportSpheres(ids=sphereIds, useRef=0,
                          what=[('rot', 'accumulatedRotation(b.id)')])


####  Particle Data Tracking  ####
def particleTracker():

    if clumpSample:
        vtkPart.exportSpheres(ids=clumpMemberIds, useRef=True,
                              what=[('dspl', 'b.state.displ()'),    # save displacement vectors

                                    # tweak rotation data to allow for using log scale in paraview
                                    #('rot', 'b.state.rot().norm()*100 + 1'),  # OLD implementation
                                    ('rotAccu', 'accumulatedRotation(b.id)'),  # accumulated rotations throughout simulation
                                    ('rot', 'b.state.rot().norm()'),    # rotation inbetween strain increments

                                    #('contacts', 'len(b.intrs())'), # number of contacts for each particle
                                    ('contacts', 'deltaContactNum(b.id)'), # change in contacts from initial state
                                    ('stepContacts', 'deltaContactNumEachStep(b.id)'),  # change in contacts from one step to the next
                                    ])
    else:
        vtkPart.exportSpheres(ids=sphereIds, useRef=True,
                              what=[('dspl', 'b.state.displ()'),
                                    ('rotAccu', 'accumulatedRotation(b.id)'),
                                    ('rot', 'b.state.rot().norm()'),
                                    ('contacts', 'deltaContactNum(b.id)'),
                                    ('stepContacts', 'deltaContactNumEachStep(b.id)'),
                                    ])

    #  useRef=True, # saves reference position of particles (b.state.refPos), if false: export current sphere position
    #  what=[("dspl","b.state.displ()")] # saves displacement vector, labeled 'dspl' in Paraview
    #  .displ() - gives displ of particle from reference pos: O.bodies[id].state.refPos

# called by strainTracker()
def paraviewData():
        #rotationTracker()  # included in displacementTracker()
        particleTracker()                       # added rotation tracking!
        forceChains()                               # normal and shear force
        #contactNumberTracker()                      # contact number for each particle
        if tesselatedMicroStrain:
            tesselatedStrain()

        if captureLocalPorosityData:
            localisedPorosity()
        #print ("res 200: ", samplePorosity(200))
        #print ("res 500: ", samplePorosity(500))

        # reset reference positions after saving data
        for i in sphereIds:  # sphereIds includes clumps and clumpMembers if clumpSample
            s = O.bodies[i]
            s.state.refPos = s.state.pos
            s.state.refOri = s.state.ori
        print("vtk-data saved")


###################################
#####  Membrane creation  #####
###################################

color1=[255./255.,102./255.,0./255.]
color2=[0,0,0]
color3=[0.5,0.5,0.5]

# Preallocate (used for both membrane types)
nodeIdsXNeg = []
nodeIdsXPos = []
nodeIdsYNeg = []
nodeIdsYPos = []

# Preallocate
connIdsXNeg = []
connIdsXPos = []
connIdsYNeg = []
connIdsYPos = []

# if necessary to differentiate between membranes for the pfacets
pFacetXNeg = []
pFacetXPos = []
pFacetYNeg = []
pFacetYPos = []
pFacetIds = [pFacetXNeg, pFacetXPos, pFacetYNeg, pFacetYPos]

nodeIds = [nodeIdsXNeg, nodeIdsXPos, nodeIdsYNeg, nodeIdsYPos]
connIds = [connIdsXNeg, connIdsXPos, connIdsYNeg, connIdsYPos]

heightM = height - 2 * radiusGN
pFacetHeight=heightM/(nh-1)  # height of each membrane square/pfacet

## separate, shifted triangular membrane grids
if genMembrane == 3:
    heightM+=membrVertExt*2    # take vertical extension of membrane into account
    gridRowH = heightM / (nh - 1)  # nh-1 shifted nodes in-between all nh nodes
    gridColW = widthM1 / (nw - 1)
    # lists for sorting nodes to make connecting simpler
    nShifted=[[],[]]
    nAligned=[[],[]]

    # create membrane nodes
    for i in range(2):  # two opposing membranes
        for h in range(nh):  # along height
            for w in range(nw):  # along width
                ## aligned nodes
                nodeIds[i].append(O.bodies.append(gridNode(
                    [wallStartPos[i] - radiusGN * pow(-1, i),
                     w * widthM1 / float(nw - 1) - membraneInc / 2 * widthY + wallStartPos[2],
                     h * heightM / float(nh - 1) + radiusGN + wallStartPos[4]-membrVertExt],
                    radiusGN, fixed=0, material='gridNodeMat', color=color1)))
                nAligned[i].append(nodeIds[i][-1])


            if h==nh-1:  # always end membrane with aligned nodes
                continue
            for w in range(nw-1):
            ## shifted nodes
                nodeIds[i].append(O.bodies.append(gridNode(
                    [wallStartPos[i] - radiusGN * pow(-1, i),
                     w*widthM1/float(nw-1)-membraneInc/2*widthY + wallStartPos[2] + gridColW/2,
                     h * heightM / float(nh - 1) + radiusGN + wallStartPos[4] + gridRowH/2-membrVertExt],
                    radiusGN, fixed=0, material='gridNodeMat', color=color1)))
                nShifted[i].append(nodeIds[i][-1])
    # connections
    for i in range(2):
        # aligned nodes
        for h in range(nh):
            for w in range(nw-1):
                # horisontal
                connIds[i].append(O.bodies.append(gridConnection(
                    nAligned[i][w + h * nw], nAligned[i][w + h * nw + 1], radiusGN, material='pFacetMat',
                    color=color3)))

        # shifted nodes
        for h in range(nh-1):
            for w in range(nw-1):
                # horisontal
                if w==nw-2:
                    pass
                else:
                    connIds[i].append(O.bodies.append(gridConnection(
                        nShifted[i][w + h * (nw-1)], nShifted[i][w + h * (nw-1) + 1], radiusGN, material='pFacetMat',
                        color=color3)))
                # triangles down
                connIds[i].append(O.bodies.append(gridConnection(
                    nShifted[i][w + h * (nw - 1)], nAligned[i][w + h * nw], radiusGN, material='pFacetMat',
                    color=color3)))
                connIds[i].append(O.bodies.append(gridConnection(
                    nShifted[i][w + h * (nw - 1)], nAligned[i][w + h * nw + 1], radiusGN, material='pFacetMat',
                    color=color3)))
                # triangles up
                connIds[i].append(O.bodies.append(gridConnection(
                    nShifted[i][w + h * (nw - 1)], nAligned[i][w + (h+1) * nw], radiusGN, material='pFacetMat',
                    color=color3)))
                connIds[i].append(O.bodies.append(gridConnection(
                    nShifted[i][w + h * (nw - 1)], nAligned[i][w + (h+1) * nw + 1], radiusGN, material='pFacetMat',
                    color=color3)))
    if genPFacets:
        # PFacets
        for i in range(2):
            for h in range(nh - 1):
                for w in range(nw - 1):
                    # shifted triangles
                    if w == nw - 2:
                        pass
                    else:
                        # shifted triangle down
                        pFacetIds[i].append(O.bodies.append(pfacet(
                            nShifted[i][w + h * (nw - 1)], nShifted[i][w + 1 + h * (nw - 1)],
                            nAligned[i][w + h * nw + 1],
                            material='pFacetMat', color=color3)))
                        # shifted triangle up
                        pFacetIds[i].append(O.bodies.append(pfacet(
                            nShifted[i][w + h * (nw - 1)], nShifted[i][w + 1 + h * (nw - 1)],
                            nAligned[i][w + (h+1) * nw + 1],
                            material='pFacetMat', color=color3)))

                    # triangle down
                    pFacetIds[i].append(O.bodies.append(pfacet(
                        nShifted[i][w + h * (nw - 1)], nAligned[i][w + h * nw], nAligned[i][w + h * nw + 1],
                        material='pFacetMat',color=color3)))
                    # triangle up
                    pFacetIds[i].append(O.bodies.append(pfacet(
                        nShifted[i][w + h * (nw - 1)], nAligned[i][w + (h+1) * nw], nAligned[i][w + (h+1) * nw + 1],
                        material='pFacetMat', color=color3)))


### 4 separate membranes
if genMembrane == 1:
    # Create nodes in membranes parallel to y-axis
    for i in range(2):          # the two membranes
        for h in range(nh):     # along height
            for w in range(nw): # along width
                nodeIds[i].append(O.bodies.append(gridNode(
                    [wallStartPos[i]-radiusGN*pow(-1,i),(w * widthM1 / float(nw - 1) - membraneInc / 2 * widthY) * pow(-1, i) +
                     widthY * i+wallStartPos[2], h * heightM / float(nh - 1)+radiusGN+wallStartPos[4]],
                     radiusGN, fixed=0, material='gridNodeMat', color=color1)))
                # dynamic is deprecated->use fixed instead

                # with diag connection mirrored -> [i * width, w * widthM / float(nw - 1) - membraneInc / 2 * width,
                #                                   h * height / float(nh - 1)],

    if not rigidWalls and numMembranes==4:
        # create nodes in membranes parallel to x-axis
        for i in range(2):
            for h in range(nh):
                for w in range(nw):
                    nodeIds[i+2].append(O.bodies.append(gridNode(
                        [(w * widthM2 / float(nw - 1) - membraneInc / 2 * widthX)* pow(-1, i) +widthX*i+wallStartPos[0],
                         wallStartPos[i+2]-radiusGN*pow(-1,i), h * heightM / float(nh - 1)+radiusGN+wallStartPos[4]],
                        radiusGN, fixed=0, material='gridNodeMat', color=color1)))

                        ###### Evt pfacetCreator3 ? -> only need 3 gridNodes as input
    # connections
    for i in range(numMembranes):
        for h in range(nh-1):
            for w in range(nw):
                ########################### Sjekke disse mot script gamle pcen
                if w >= (nw-1):
                    connIds[i].append(O.bodies.append(gridConnection(
                        nodeIds[i][w + h * nw ], nodeIds[i][w + (h+1) * nw], radiusGN, material='pFacetMat',color=color3)))
                    continue
                if h >= nh-2 and w < nw-1:
                    connIds[i].append(O.bodies.append(gridConnection(
                        nodeIds[i][w + (h+1) * nw], nodeIds[i][w + (h+1) * nw + 1], radiusGN, material='pFacetMat',
                                       color=color3)))
                # horisontal connection
                connIds[i].append(O.bodies.append(gridConnection(
                    nodeIds[i][w+h*nw], nodeIds[i][w+h*nw+1], radiusGN, material='pFacetMat', color=color3)))
                # vertical connection
                connIds[i].append(O.bodies.append(gridConnection(
                    nodeIds[i][w+h*nw], nodeIds[i][w+h*nw+nw], radiusGN, material='pFacetMat', color=color3)))
                if genPFacets:
                    # diagonal connection
                    connIds[i].append(O.bodies.append(gridConnection(
                        nodeIds[i][w+h*nw], nodeIds[i][w+h*nw+nw+1], radiusGN, material='pFacetMat', color=color3)))
    if genPFacets:
        # PFacets
        for i in range(numMembranes):
            for n in range(0,len(nodeIds[i])-nw):
                if (n+1)%nw == 0:
                    continue
                pFacetIds[i].append(O.bodies.append(
                    pfacet(nodeIds[i][n], nodeIds[i][n+1],nodeIds[i][n+nw+1], material='pFacetMat', color=color3)))
                pFacetIds[i].append(O.bodies.append(
                    pfacet(nodeIds[i][n], nodeIds[i][n + nw], nodeIds[i][n + nw + 1], material='pFacetMat', color=color3)))


#########################################################################################
## Continuous membrane
if genMembrane == 2:
    #widthM = width + 2 * radiusGN
    ## Grid nodes
    for h in range(nh):  # along height
        for i in range(2):  # the two sides normal to x-axis
            for w in range(nw):  # along width
                nodeIds[i].append(O.bodies.append(gridNode(
                    [i * width - radiusGN * pow(-1, i), w * widthM1 /float(nw - 1)-radiusGN,
                     h * heightM / float(nh - 1)+radiusGN],
                    radiusGN, fixed=0, material='gridNodeMat',color=color1)))

        for i in range(2):
            nodeIds[i + 2].append(nodeIds[0][0 + h * nw - i*(1+nw*h)]) # same node as in first membrane
            for w in range(1,nw-1): # all nodes in between the corner nodes
                nodeIds[i+2].append(O.bodies.append(gridNode(
                    [w * widthM2 /float(nw - 1)-radiusGN, i * width - radiusGN * pow(-1, i),
                     h * heightM / float(nh - 1)+radiusGN],
                    radiusGN, fixed=0, material='gridNodeMat', color=color1)))
            nodeIds[i + 2].append(nodeIds[1][0 + h * nw - i*(1+nw*h)]) # same node as in second membrane

    ## Connections
    for i in range(2):
        for h in range(nh - 1):
            for n in range(nw - 1):
                # horisontal
                connIds[i].append(O.bodies.append(
                    gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + 1],radiusGN,
                                   material='pFacetMat', color=color3)))
                # vertical
                connIds[i].append(O.bodies.append(
                    gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + nw], radiusGN,
                                   material='pFacetMat', color=color3)))
                if genPFacets:
                    # diagonal
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + nw + 1], radiusGN,
                                       material='pFacetMat', color=color3)))
                # last con over height
                if n == nw-2:
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n+1 + h * nw], nodeIds[i][n+1 + h * nw + nw], radiusGN,
                                       material='pFacetMat', color=color3)))
                if h == nh-2:
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n + (h+1) * nw], nodeIds[i][n + (h+1) * nw + 1],radiusGN,
                                       material='pFacetMat', color=color3)))
    for i in range(2,4):
        for h in range(nh - 1):
            for n in range(nw - 1):
                # horisontal
                connIds[i].append(O.bodies.append(
                    gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + 1], radiusGN,
                                   material='pFacetMat', color=color3)))
                if genPFacets:
                    # diagonal
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + nw + 1], radiusGN,
                                       material='pFacetMat', color=color3)))
                if h == nh-2:
                    # top horisontal
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n + (h+1) * nw], nodeIds[i][n + (h+1) * nw + 1], radiusGN,
                                       material='pFacetMat', color=color3)))
                if n>0 and n<nw-1:
                    # vertical
                    connIds[i].append(O.bodies.append(
                        gridConnection(nodeIds[i][n + h * nw], nodeIds[i][n + h * nw + nw], radiusGN,
                                       material='pFacetMat', color=color3)))
    if genPFacets:
        ## Pfacets
        for i in range(4):
            for h in range(nh - 1):
                for w in range(nw - 1):
                    # bottom triangle
                    pFacetIds.append(O.bodies.append(
                        pfacet(nodeIds[i][w + h*nw], nodeIds[i][w + 1 + h*nw], nodeIds[i][w + nw + 1 + h*nw],
                               material='pFacetMat', color=color3)))
                    # top triangle
                    pFacetIds.append(O.bodies.append(
                        pfacet(nodeIds[i][w + h*nw], nodeIds[i][w + nw + h*nw], nodeIds[i][w + nw + 1 + h*nw],
                               material='pFacetMat', color=color3)))


if genMembrane==1 or genMembrane==2:
    # constraining z-dof of the nodes
    if verticalLock:
        for i in range(numMembranes):
            for j in nodeIds[i]:
              O.bodies[j].state.blockedDOFs='z'

    if parallelLock:
        # constrain dof parallel to membrane
        for i in range(0,2):
            for j in nodeIds[i]:
                O.bodies[j].state.blockedDOFs = 'yz'
        for i in range(2,4):
            for j in nodeIds[i]:
                O.bodies[j].state.blockedDOFs = 'xz'


##############################
######  Cover PFacets  ######
##############################
if genCovFacets!=0 and genMembrane!=0:
    covFric=0  # default setting -> zero friction with membrane
    if coverFriction:
        covFric=finalFricDegree
    O.materials.append(CohFrictMat(young=membraneStiff, poisson=.3, density=densityGridNode, frictionAngle=radians(covFric),
                                   normalCohesion=3e20, shearCohesion=3e20, momentRotationLaw=momRotLaw,
                                   label='coverNodeMat'))
    O.materials.append(FrictMat(young=youngFacet * 10e7, poisson=.3, density=densityFacet, frictionAngle=radians(covFric),
                                label='sideCovFacetMat'))
    O.materials.append(
        FrictMat(young=youngFacet * 10e7, poisson=.3, density=densityFacet, frictionAngle=radians(0),
                 label='vertCovFacetMat'))
    coverFixed = 1
    covNodes = []
    covConn = []
    covFacets = []
    covRad = widthX*0.05

    covInc = membraneInc * 1.
    covW = width * (1 + 0.5 * covInc)
    covH = height * 0.5
    covInto = -covRad

# connection based covers
if genCovFacets==2 and genMembrane!=0:
    radDist = radiusGN + covRad  # rad of membrane and covers combined
    avgCovRad = (radiusGN + covRad) / 2.
    covSideH = height * 0.3
    covSideT = membraneInc / 2 * width  # thickness
    covSideMemW = widthY * membraneInc / 2 - covRad  # - radDist - avgCovRad
    covSideExt = width * 0.3  # extend covers this dist outside the membrane
    #if shallowSample: covSideExt = width*1.2
    covSideW = covSideMemW+covSideExt
    topCoverNodeIds = []  # containing just top covers -> to prescribe velocity to these plates

    cnw = 14  # num connections along each side cover
    if shallowSample:
        cnw = 9
    cnh = 14  # num connections along height of top and bot cover

    conCovGN=[[],[]]
    conCovCon=[[],[]]
    conCovFac=[[],[]]
    topCoverNodeIds=[]
    botCoverNodeIds = []
    sideCoverNodeIds=[[0,0],[0,0]]

    if genSideCovFacets:
        ## Side covers
        for i in range(2):  #Xmin and Xmax -> two separate membranes
            for j in range(2):  #Ymin and Ymax
                for h in range(2):
                    for w in range(cnw):
                        conCovGN[i].append(O.bodies.append(
                            gridNode([wallStartPos[i]+covRad*pow(-1,i),
                                      wallStartPos[2+j]-(covRad+w*covSideW/float(cnw-1))*pow(-1,j),
                                      wallStartPos[4+h] + covRad*pow(-1,h)],
                                     covRad, fixed=coverFixed, material='coverNodeMat', color=color1)))
                ## connections
                for n1,n2 in zip(conCovGN[i][-2*cnw:-cnw],conCovGN[i][-cnw:]):
                    conCovCon[i].append(O.bodies.append(gridConnection(
                                        n1, n2, covRad, material='sideCovFacetMat', color=color1)))
        # xmin: ymin, ymax
        sideCoverNodeIds[0][0] = conCovGN[0][-4*cnw:-2*cnw]
        sideCoverNodeIds[0][1] = conCovGN[0][-2 * cnw:]
        #xmax: ymin, ymax
        sideCoverNodeIds[1][0] = conCovGN[1][-4 * cnw:-2 * cnw]
        sideCoverNodeIds[1][1] = conCovGN[1][-2 * cnw:]

    if genTopBotCovFacets:
        # Top and bot covers
        for i in range(2):  #Xmin and Xmax -> two separate membranes
            for j in range(2): # bottom, then top
                for w in range(2):  # Ymin and Ymax
                    for h in range(cnh):
                        conCovGN[i].append(O.bodies.append(
                            gridNode([wallStartPos[i]+covRad*pow(-1,i),
                                      wallStartPos[2+w]-(covRad+covSideW)*pow(-1,w),
                                      wallStartPos[4+j]-(covRad+h*covSideH/float(cnh))*pow(-1,j)],
                                     covRad, fixed=coverFixed, material='coverNodeMat', color=color1)))
                ## connections
                for n1,n2 in zip(conCovGN[i][-2*cnh:-cnh],conCovGN[i][-cnh:]):
                    conCovCon[i].append(O.bodies.append(gridConnection(
                                        n1, n2, covRad, material='sideCovFacetMat', color=color1)))

        topCoverNodeIds=conCovGN[0][-2*cnh:] + conCovGN[1][-2*cnh:]
        botCoverNodeIds=conCovGN[0][-4*cnh:-2*cnh] + conCovGN[1][-4*cnh:-2*cnh]

    # add all cover nodes for designating material masks
    covNodes=conCovGN[0]+conCovGN[1]
    covConn=conCovCon[0]+conCovCon[1]

if genCovFacets==1 and genMembrane!=0:
    ### Create covering plates to block membrane

    if genFullBlock:
        gap = (radiusGN+covRad)
        for i in range(2):

            n1 = O.bodies[nodeIds[i][0]].state.pos + Vector3(gap*pow(-1,i), 0, 0)
            n2 = O.bodies[nodeIds[i][nw - 1]].state.pos + Vector3(gap*pow(-1,i), 0, 0)
            n3 = O.bodies[nodeIds[i][-nw]].state.pos + Vector3(gap*pow(-1,i), 0, 0)
            vertices = [n1, n2, n3]
            pfacetCreator1(vertices, covRad, nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                           fixed=coverFixed, materialNodes='coverNodeMat', material='sideCovFacetMat')
            # pfacetCreator1(vertices,r,nodesIds=[],cylIds=[],pfIds=[],wire=False,color=color,
            # fixed=fixed,materialNodes='gridNodeMat',material='gridConnectionMat')
            n4 = O.bodies[nodeIds[i][-1]].state.pos + Vector3(gap*pow(-1,i), 0, 0)
            vertices = [n2, n3, n4]
            pfacetCreator1(vertices, covRad, nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                           fixed=coverFixed, materialNodes='coverNodeMat', material='sideCovFacetMat')
    else:
        radDist = radiusGN + covRad  # rad of membrane and covers combined
        avgCovRad = (radiusGN + covRad) / 2.
        covSideH = height * 0.7
        covSideT = membraneInc / 2 * width  # thickness
        covSideW = widthY * membraneInc / 2 - covRad #- radDist - avgCovRad
        covSideExt = width * 0.3  # extend covers this dist outside the membrane

        topCoverNodeIds=[]  # containing just top covers -> to prescribe velocity to these plates
        if genTopBotCovFacets:
            ## Top and bottom covers
            for i in range(2):  # xmin and xmax cover pairs
                ## Bottom cover
                # Ymin nodes
                n1 = Vector3(wallStartPos[i]+covRad*pow(-1,i),wallStartPos[2]-covSideW-covSideExt-covRad,
                             wallStartPos[4]-covRad)
                n2 = Vector3(wallStartPos[i]+covRad*pow(-1,i),wallStartPos[3]+covSideW+covSideExt+covRad,
                             wallStartPos[4]-covRad)
                # Ymax nodes
                n3 = n2 +Vector3(0,0,-covSideH)
                n4 = n1 +Vector3(0,0,-covSideH)
                vertices=[n1, n2, n3]
                pfacetCreator1(vertices,covRad,nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                               fixed=coverFixed, materialNodes='coverNodeMat', material='vertCovFacetMat')
                vertices = [n1, n3, n4]
                pfacetCreator1(vertices, covRad, nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                               fixed=coverFixed, materialNodes='coverNodeMat', material='vertCovFacetMat')

                ## Top cover
                # Ymin nodes
                n1 = Vector3(wallStartPos[i] + covRad * pow(-1, i), wallStartPos[2] - covSideW - covSideExt - covRad,
                             wallStartPos[5] + covRad)

                n2 = Vector3(wallStartPos[i] + covRad * pow(-1, i), wallStartPos[3] + covSideW + covSideExt + covRad,
                             wallStartPos[5] + covRad)
                # Ymax nodes
                n3 = n2 + Vector3(0, 0, +covSideH)
                n4 = n1 + Vector3(0, 0, +covSideH)
                vertices=[n1,n2,n3]
                pfacetCreator1(vertices,covRad,nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                               fixed=coverFixed, materialNodes='coverNodeMat', material='vertCovFacetMat')
                vertices = [n1, n3, n4]
                pfacetCreator1(vertices, covRad, nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                               fixed=coverFixed, materialNodes='coverNodeMat', material='vertCovFacetMat')
                # collect all top cover node ids in list
                topCoverNodeIds.append(covNodes[-4:])

        if genSideCovFacets:
            ## Side covers
            for i in range(2): # for xmin and xmax
                for j in range(2): # for ymin and ymax
                    # node positions
                    n1 = Vector3(wallStartPos[i] + covRad * pow(-1, i), wallStartPos[2+j]-covRad*pow(-1,j),
                                 wallStartPos[4] + covRad)
                    n2 = Vector3(wallStartPos[i] + covRad * pow(-1, i), wallStartPos[2+j]-covRad*pow(-1,j),
                                 wallStartPos[5] - covRad)

                    n3 = n2 + Vector3(0,-(covSideW+covSideExt)*pow(-1,j),0)
                    n4 = n1 + Vector3(0,-(covSideW+covSideExt)*pow(-1,j),0)

                    # n1 = corPos[0 + 2*j] + Vector3(radDist * pow(-1, i), covSideExt * -pow(-1, j)*pow(-1, i), 0)
                    # n2 = corPos[1 + 2*j] + Vector3(radDist * pow(-1, i), covSideExt * -pow(-1, j)*pow(-1, i), 0)
                    # n3 = corPos[0 + 2*j] + Vector3(radDist * pow(-1, i), covSideW * pow(-1, j)*pow(-1, i), 0)
                    # n4 = corPos[1 + 2*j] + Vector3(radDist * pow(-1, i), covSideW * pow(-1, j)*pow(-1, i), 0)
                    vertices=[n1,n2,n3]
                    pfacetCreator1(vertices,covRad,nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                                   fixed=coverFixed, materialNodes='coverNodeMat', material='sideCovFacetMat')
                    vertices = [n1, n3, n4]
                    pfacetCreator1(vertices, covRad, nodesIds=covNodes, cylIds=covConn, pfIds=covFacets, color=color1,
                                   fixed=coverFixed, materialNodes='coverNodeMat', material='sideCovFacetMat')

##############################
#####   Anchor nodes   #####
##############################
# initialize
anchorNodes = []
anchorConn = []
anchorFacets = []
anchorHeight = pFacetHeight
anchorRadius = radiusGN
aPos = Vector3(0,0,0)

if genMembrane == 1 and genAnchors == 1:
    # Materials for anchor elements
    O.materials.append(CohFrictMat(young=membraneStiff, poisson=.3, density=densityFacet, frictionAngle=radians(30),
                                   normalCohesion=3e20, shearCohesion=3e20, momentRotationLaw=momRotLaw,
                                   label='anchorNodeMat'))
    O.materials.append(FrictMat(young=youngFacet, poisson=.3, density=densityFacet, frictionAngle=radians(30),
                                label='anchorConnMat'))

    anchorXP = [[],[]]  # parallel to x axis, positive side
    anchorXN = [[],[]]
    anchorYP = [[],[]]
    anchorYN = [[],[]]
    anchorNodes = [anchorXN, anchorXP, anchorYN, anchorYP]
    # existing nodes to connect to. -   nw*2 for top and bottom anchors
    anchorConNodes = [[[],[]] for i in range(numMembranes)]
    ## vertical anchor nodes
    # get the bordernodes' position and lower/raise the anchornode height by height/(nh-1) -> element height
    for i in range(numMembranes):
        # bottom nodes
        for j in range(0,nw):
            n = nodeIds[i][j]
            anchorConNodes[i][0].append(n)
            aPos = O.bodies[n].state.pos
            aPos[2] -= anchorHeight
            anchorNodes[i][0].append(O.bodies.append(gridNode(aPos, anchorRadius, fixed=0, material='anchorNodeMat',
                                                           color=color1)))
        # top nodes
        for j in range(-nw,0):
            n = nodeIds[i][j]
            anchorConNodes[i][1].append(n)
            aPos = O.bodies[n].state.pos
            aPos[2]+=anchorHeight
            anchorNodes[i][1].append(O.bodies.append(gridNode(aPos,anchorRadius,fixed = 0, material = 'anchorNodeMat',
                                                           color = color1)))

    ## Vertical connections for anchors
    for i in range(numMembranes):
        # bot
        for j in range(len(anchorNodes[i][0])):
            anchorConn.append(O.bodies.append(gridConnection(anchorNodes[i][0][j], anchorConNodes[i][0][j],
                                                             anchorRadius, material = 'pFacetMat', color = color3)))
        # top
        for j in range(len(anchorNodes[i][1])):
            anchorConn.append(O.bodies.append(gridConnection(anchorNodes[i][1][j], anchorConNodes[i][1][j],
                                                             anchorRadius, material = 'pFacetMat', color = color3)))
    ## extending membrane vertically
    for i in range(numMembranes):
        for j in range(2):  # bottom and top anchors
            for h in range(extendedMembraneRows-1):
                # position list
                aPos = [O.bodies[n].state.pos - pow(-1,j)*Vector3(0,0,anchorHeight) for n in anchorNodes[i][j][-nw:]]

                for w in range(nw):
                    # grid nodes
                    anchorNodes[i][j].append(
                        O.bodies.append(gridNode(aPos[w], anchorRadius, fixed=0, material='anchorNodeMat',
                                                 color=color1)))
                # vertical connection
                    anchorConn.append(
                        O.bodies.append(gridConnection(anchorNodes[i][j][w+h*nw], anchorNodes[i][j][w+(h+1)*nw], anchorRadius,
                                                       material='pFacetMat', color=color3)))
                for w in range(nw):
                # horisontal connection
                    if w == nw-1:
                        continue
                    else:
                        anchorConn.append(O.bodies.append(gridConnection(
                            anchorNodes[i][j][w + h*nw], anchorNodes[i][j][w + h*nw + 1],anchorRadius,
                            material='pFacetMat', color=color3)))
                    if h==extendedMembraneRows-2 and w<nw-1:
                        anchorConn.append(O.bodies.append(gridConnection(
                            anchorNodes[i][j][w + (h+1) * nw], anchorNodes[i][j][w + (h+1) * nw + 1], anchorRadius,
                            material='pFacetMat', color=color3)))

    if anchorFacets:
        # horisontal and diagonal connections
        for i in range(numMembranes):
            for j in range(nw-1):
                ## top nodes
                # diagonal connection
                anchorConn.append(
                    O.bodies.append(gridConnection(anchorNodes[i][j+1], anchorConNodes[i][j], anchorRadius,
                                                                 material='pFacetMat', color=color3)))
                # horisontal connection
                anchorConn.append(
                    O.bodies.append(gridConnection(anchorNodes[i][j], anchorNodes[i][j + 1], anchorRadius,
                                                   material='pFacetMat', color=color3)))

                ## bottom nodes
                # diagonal connection
                anchorConn.append(
                    O.bodies.append(gridConnection(anchorNodes[i][j + nw], anchorConNodes[i][j + nw + 1], anchorRadius,
                                                   material='pFacetMat', color=color3)))
                # horisontal connection
                anchorConn.append(
                    O.bodies.append(gridConnection(anchorNodes[i][j + nw], anchorNodes[i][j + 1 + nw], anchorRadius,
                                                   material='pFacetMat', color=color3)))

                ## pfacet
                # top
                anchorFacets.append(O.bodies.append(
                    pfacet(anchorConNodes[i][j], anchorNodes[i][j], anchorNodes[i][j + 1], material='pFacetMat', color=color3)))
                anchorFacets.append(O.bodies.append(
                    pfacet(anchorConNodes[i][j], anchorConNodes[i][j+1], anchorNodes[i][j+1], material='pFacetMat', color=color3)))
                # bottom
                anchorFacets.append(O.bodies.append(
                    pfacet(anchorNodes[i][j+nw], anchorConNodes[i][j+nw], anchorConNodes[i][j + 1+nw], material='pFacetMat',
                           color=color3)))
                anchorFacets.append(O.bodies.append(
                    pfacet(anchorNodes[i][j+nw], anchorNodes[i][j + 1+nw], anchorConNodes[i][j + 1+nw], material='pFacetMat',
                           color=color3)))


    if lockAnchors:
        # constraining all dofs of the nodes
        for twoList in anchorNodes:
            for nList in twoList:
                for n in nList:
                  O.bodies[n].state.blockedDOFs='xyz'
    if lockAnchorsZ:
        # constraining anchorNodes in z-dir
        for twoList in anchorNodes:
            for nList in twoList:
                for n in nList:
                    O.bodies[n].state.blockedDOFs = 'z'

    if lockTopAnchorsZ:
        for twoList in anchorNodes:
                for n in twoList[1][-nw:]:  # [1] - top nodes
                    O.bodies[n].state.blockedDOFs = 'z'

    if lockTopAnchorsX:
        for twoList in anchorNodes:
            for n in twoList[1][-nw:]:  # [1] - top nodes
                    O.bodies[n].state.blockedDOFs = 'x'

##############################
#####   Material Masks   #####
##############################

#### Change all loaded spheres material mask to 11 in order to interact with the walls and both membranes (mask 1 and 2), while the
# membranes cant interact with each other nor the walls.
## Masks:
# particles:11=1011       mem1:5=0101   mem2:2=0010   covers:4=0100    walls:8=1000

particleMask = 11  # 1011
for s in sphereIds:
    O.bodies[s].mask=particleMask

if rigidWalls!=0:
    wallMask = 8  # 1000
    for w in wallBodies:
        w.mask=wallMask


## Inhibit interactions between overlapping membranes by groupmask
if genMembrane!=0:
    maskM1 = 5  # binary -> 0101   -> Also interacts with covFacets mask = 4 - 0100
    maskM2 = 2  # binary -> 0010
    # mask for particles is set to mask=3 -> 0011 will interact with both mask=1 and mask=2 as they have a common
    # active bit, but 1 and 2 will not interact with each other.

    ## membranes parallel to y-axis
    for i in range(0, 2):
        # gridnodes
        for n in nodeIds[i]:
            O.bodies[n].mask = maskM1
        # connections
        for c in connIds[i]:
            O.bodies[c].mask = maskM1
        # P-facets
        for p in pFacetIds[i]:
            O.bodies[p].mask = maskM1
    ## membranes parallel to x-axis
    for i in range(2, 4):
        for n in nodeIds[i]:
            O.bodies[n].mask = maskM2
        for c in connIds[i]:
            O.bodies[c].mask = maskM2
        for p in pFacetIds[i]:
            O.bodies[p].mask = maskM2


## coverFacets
if genCovFacets and genMembrane!=0:
    ## Inhibit interactions with sample
    covMask = 4  # binary -> 0100
    # mask for particles is set to mask=11 -> 1011. Membrane 1 set to mask=5 -> 0101

    for i in range(0,2):
        #gridnodes
        for n in covNodes:
            O.bodies[n].mask = covMask
        # connections
        for c in covConn:
            O.bodies[c].mask = covMask
        # P-facets
        for p in covFacets:
            O.bodies[p].mask = covMask
    # ## Make covFacets interact with parallel membrane
    # covMask = 2  # binary -> 0001
    # maskM2 = 2  # binary -> 0010
    # # mask for particles is set to mask=3 -> 0011 will interact with both mask=1 and mask=2 as they have a common
    # # active bit, but 1 and 2 will not interact with each other.

## anchor interactions
if genAnchors:
    # Same mask as the connected membrane
    for i in range(numMembranes):
        for j in range(2):  # top and bottom anchors
            #gridnodes
            for n in anchorNodes[i][j]:
                O.bodies[n].mask = maskM1
    # connections
    for c in anchorConn:
        O.bodies[c].mask = maskM1
    # P-facets
    for p in anchorFacets:
        O.bodies[p].mask = maskM1

## binary
# 1 - 0001
# 2 - 0010
# 3 - 0011
# 4 - 0100
# 5 - 0101
# 6 - 0110
# 7 - 0111
# 8 - 1000
# 9 - 1001

# 15- 1111

###################################
#####   Load plates   #####
###################################
# add bottom plate as facet?
#O.bodies.append(geom.facetBox((0,0,.1),(.12,.12,.12),wallMask=16))  # wallMask 16 gives only -z wall of box (binary)


# ######======= BEFORE IMPORTING FACETS FROM SAMPLEGEN  -> ALL OF THIS WAS USED  ##################
# # or add bottom plate as wall-element
# O.bodies.append(wall(wallStartPos[4],axis=2,sense=1,material='walls'))
# bottomPlate=O.bodies[-1]
#
#
# # add top plate
# O.bodies.append(wall(wallStartPos[5], axis=2, sense=-1, material='walls'))
# global plate  # without this line, the plate variable would only exist inside this function
# topPlate = O.bodies[-1]  # the last body is the plate
# # start height of plate
# plateStartZ = topPlate.state.pos[2]
#
#
# # the maximum finds the z-coordinate of the top of the topmost particle
# # O.bodies.append(wall(max([b.state.pos[2]+b.shape.radius for b in O.bodies if isinstance(b.shape,Sphere)]),axis=2,sense=-1))
#
# if rigidWalls!=0:
#     ## add supporting walls
#     O.bodies.append(wall(wallStartPos[2], axis=1, sense=1, material='supWalls'))
#     wallYMin=O.bodies[-1]
#     O.bodies.append(wall(wallStartPos[3], axis=1, sense=-1, material='supWalls'))
#     wallYMax=O.bodies[-1]
#
# if rigidWalls==2:
#     ## add supporting walls instead of membrane
#     O.bodies.append(wall(wallStartPos[0], axis=0, sense=1, material='supWalls'))
#     wallXMin = O.bodies[-1]
#     O.bodies.append(wall(wallStartPos[1], axis=0, sense=-1, material='supWalls'))
#     wallXMax = O.bodies[-1]
#
# ######=======                                                                    ##################

############################################
#####   Apply lateral stress   #####
############################################

# for determination of directional vector
sampleCentre = (width/2, width/2, height/2)

# area and lumped forces
area = width*height  # total initial surface area of 1 membrane in contact with sample
numPFacets = ((nw-1)*(nh-1))*2  # 2 pfacet triangles pr square
areaPF = area/numPFacets
# for each pfacet triangle element
supportForce = -sigmaIso*areaPF

#areaM = widthM*heightM  # TOTAL AREA OF MEMBRANE for initial sigmaX=sigmaY
areaM1 = widthM1*heightM  # total membrane area in the yz plane
areaM2 = widthM2*heightM  #total membrane area in the xz plane

totalForceM1 = -sigmaIso*areaM1
totalForceM2 = -sigmaIso*areaM2

# direction perpendicular to each respective membrane
dirVector = [Vector3(1,0,0), Vector3(-1,0,0), Vector3(0,1,0), Vector3(0,-1,0)]
forceVector = [n*supportForce for n in dirVector]

forceVector2 = [2*n*supportForce for n in dirVector]

supportForceHalf = supportForce*0.5

if lateralStress==1:
    ## load func for membrane consisting of horisontal strips
    if nw==2 and genMembrane!=0:
        # Preallocate
        internalNodes = []
        bottomNodes = []
        topNodes = []
        ## Sort internal nodes and border nodes
        for i in range(numMembranes):
            internalNodes.append(nodeIds[i][nw:-2 * nw])
            bottomNodes.append(nodeIds[i][0:nw])
            topNodes.append(nodeIds[i][-2 * nw:])

        def confiningPres():
            ## Check/change top nodes dep on wall displ. The pos of node topNodes[0][0] is used to check against wall pos as
            # All nodes' z-dof is locked
            if topPlate.state.pos[2] < O.bodies[topNodes[0][0]].state.pos[2]:
                for i in range(numMembranes):
                    # assign new top nodes and remove them from internalNodes
                    topNodes[i][2], topNodes[i][3] = topNodes[i][0],topNodes[i][1]
                    topNodes[i][0:nw] = internalNodes[i][-nw:]
                    del internalNodes[i][-nw:]
            ## Top nodes - varying force dep on wall displ
            # Update top node loads
            updateReducedLoad()
            for i in range(numMembranes):
                # apply force on the nodes
                for j in range(len(topNodes[i])):
                    n = topNodes[i][j]
                    O.forces.addF(n, reducedForceTop[j]*dirVector[i])
                # internal nodes
                for n in internalNodes[i]:
                    O.forces.addF(n, supportForce * dirVector[i])
                # bottom nodes
                for j in range(len(bottomNodes[i])):
                    n = bottomNodes[i][j]
                    O.forces.addF(n, reducedForceBot[j] * dirVector[i])


    ## load func for membrane of more elements in the horisontal dir
    if nw>2 and genMembrane!=0:
        # Preallocate
        topNodes = []
        bottomNodes = []
        internalNodes = [[], [], [], []]
        sideNodesVert = [[], [], [], []]
        # horis sideNodes are just in topNodes and bottomNodes ofc

        ## Sort nodes
        for i in range(numMembranes):
            topNodes.append(nodeIds[i][-2 * nw:])  # two top rows
            bottomNodes.append(nodeIds[i][0:nw])  # bottom row
            for h in range(1, nh - 2):  # the two top rows are for topNodes and the bottom row is for bottomNodes
                internalNodes[i] += nodeIds[i][1 + h * nw:nw - 1 + h * nw]
                sideNodesVert[i] += [nodeIds[i][0 + h * nw], nodeIds[i][nw-1 + h * nw]]  # first and last node of each row

        supForceInt = 2 * supportForce  # 1/3 of 6 triangles
        supForceSideVert = supportForce  # 1/3 of 3 triangles # side nodes over vertical direction
        supForceSideHoris = supForceSideVert  # side nodes stretching over horisontal dir (top and bottom of membrane)
        supForceCor1 = supForceCor2 =supForceCor3 =supForceCor4 = 0.5 * supportForce  # cornerNodes -> 1/2 of sideNodes

        if massDistributedLoad: # forces dep. on mass of each node -> Remember: mass from connections is lumped in nodes!
            totalMemMass=0
            for n in nodeIds[0]:
                totalMemMass+=O.bodies[n].state.mass
            supForceInt = O.bodies[nodeIds[0][1+nw]].state.mass/totalMemMass*totalForceM1     # 2. row and 2. column from bot
            supForceSideVert = O.bodies[nodeIds[0][nw]].state.mass/totalMemMass*totalForceM1  # 2. row, 1. node
            supForceSideHoris = O.bodies[nodeIds[0][1]].state.mass/totalMemMass*totalForceM1  # 2. node from bot
            supForceCor1 = O.bodies[nodeIds[0][0]].state.mass / totalMemMass * totalForceM1
            supForceCor2 = O.bodies[nodeIds[0][nw-1]].state.mass / totalMemMass * totalForceM1
            supForceCor3 = O.bodies[nodeIds[0][-nw]].state.mass / totalMemMass * totalForceM1
            supForceCor4 = O.bodies[nodeIds[0][-1]].state.mass / totalMemMass * totalForceM1


            # forceVectors for upper membrane
        # second top row
        supForceTopList = [supForceSideVert] + [supForceInt] * (nw - 2) + [supForceSideVert]
        # top row
        supForceTopList += [supForceCor3] + [supForceSideHoris] * (nw - 2) + [supForceCor4]
        # bottom row
        supForceBotList = [supForceCor1] + [supForceSideHoris] * (nw - 2) + [supForceCor2]
        supForceTopListStart = supForceTopList


        def confiningPres():
        # change top nodes
            #global supForceTopList
            if topPlate.state.pos[2] < O.bodies[topNodes[0][0]].state.pos[2]:

                #supForceTopList = supForceTopListStart  # reset supForceTopList for the new top nodes
                for topN,intN in zip(topNodes,internalNodes):  # iterate over both list simultaneously
                    # assign new top nodes and remove them from internalNodes
                    topN[-nw:]=topN[:nw]    # move bottom row of top nodes up
                    topN[:nw]=intN[-nw:]    # copy nodes from int to top
                    del intN[-nw:]          # delete copied internal nodes


            # update forces on top nodes dep on vertical displ of wall
            updateReducedLoadMoreNW()
            for j in range(nw, 2*nw):
                supForceTopList[j]=supForceTopList[j]* reductionTopRow  # First row -> factor 1 to 0
            for j in range(nw):
                supForceTopList[j] = supForceTopList[j] * reductionSecRow  # Second row -> factor 1 to 1/2

            ## apply forces
            for i in range(numMembranes):
                # internal nodes
                for n in internalNodes[i]:
                    O.forces.addF(n, supForceInt*dirVector[i])
                # side nodes
                for n in sideNodesVert[i]:
                    O.forces.addF(n, supForceSideVert*dirVector[i])
                # top nodes - 2 rows
                for n, nodeForce in zip(topNodes[i], supForceTopList):
                    O.forces.addF(n, nodeForce * dirVector[i])
                # bottom nodes - 1 row
                for n, nodeForce in zip(bottomNodes[i], supForceBotList):
                    O.forces.addF(n, nodeForce * dirVector[i])


if lateralStress==2:
    totalMem1Mass=0

    # Find total mass of membrane + anchors
    for n in nodeIds[0]:
        totalMem1Mass +=O.bodies[n].state.mass
    print ("Mass active memb = ",totalMem1Mass)
    if genAnchors:
        for nTB in anchorNodes[0]:  # top and bot separated into two lists
            for n in nTB:
                totalMem1Mass += O.bodies[n].state.mass
        print("Mass memb incl anchors = ", totalMem1Mass)

    def confiningPresAll():  ## apply the force
        for i in range(numMembranes):
            for n in nodeIds[i]:
                massFraction = O.bodies[n].state.mass/totalMem1Mass
                O.forces.addF(n, massFraction * totalForceM1 * dirVector[i])
            if genAnchors:
                for nTB in anchorNodes[i]:  # top and bot
                    for n in nTB:
                        massFraction = O.bodies[n].state.mass/totalMem1Mass
                        O.forces.addF(n, massFraction * totalForceM1 * dirVector[i])


activeNodes = {}
forceControlList=[[],[]]

if lateralStress==3:
    nodeOutsideDist={}  # preallocate
    #=#= sort nodeIds into list of connected triangles
    nodeTriangles=[[],[]]
    intrsIds = []  # momentary id list
    #intrsIds2 = []
    for mem in range(numMembranes):
        for n in nodeIds[mem]:
            for i in O.bodies[n].intrs():
                # find connected nodes
                if i.id1!=n: intrsIds.append(i.id1)
                else: intrsIds.append(i.id2)
            for n2 in intrsIds:
                for i in O.bodies[n2].intrs():
                    # check that not node n, not node n2, but connected to both of these
                    if i.id1 != n and i.id1!=n2 and i.id1 in intrsIds:
                        tri = [n,n2,i.id1]
                        tri.sort()
                        if tri not in nodeTriangles[mem]:
                            nodeTriangles[mem].append(tri)
                    if i.id2 != n and i.id2 != n2 and i.id2 in intrsIds:
                        tri = [n, n2, i.id2]
                        tri.sort()
                        if tri not in nodeTriangles[mem]:
                            nodeTriangles[mem].append(tri)
            intrsIds=[]  # reset list
    #=#=

    # REMEMBER TO SET CORRECT CONFINING-ENGINE WHEN lateralStress==3
    activeTotalForce=[0,0]
    def confiningPres():
        #nodeActive={}   # initialize bool dict - True/False
        global activeTotalForce
        activeTotalForce = [0, 0]
        global active1Triangles
        global active2Triangles
        global active3Triangles
        # reset triangle selection
        active1Triangles=[[],[]]
        active2Triangles=[[],[]]
        active3Triangles=[[],[]]
        global nodeOutsideDist
        nodeOutsideDist={}  # dict containing dist to rigid walls from outside node

        for i in range(numMembranes):
            for t in nodeTriangles[i]:
                num=numActiveNodes(i, t)

                if num==0:
                    continue
                elif num==1:  # 1 node of triangle in active area
                    active1Triangles[i].append(t)
                elif num==2:  # 2 nodes of triangle in active area
                    active2Triangles[i].append(t)
                elif num==3:  # whole triangle in active area
                    active3Triangles[i].append(t)


            for t in active3Triangles[i]:
                if pressureNormToMem:
                    crossVec = calc3TriArea(i,t)
                    for n in t:
                        O.forces.addF(n, sigmaX*crossVec*1/3.)  # crossvec points out from the surface. Need minus sigmaX in pressure
                        activeTotalForce[i]+=-sigmaX*crossVec.norm()*1/3.
                else:
                    nodeForce = -sigmaX*calc3TriArea(t) * 1/3.  # 1/3 for each node in the triangle
                    for n in t:
                        O.forces.addF(n, nodeForce* dirVector[i])
                        activeTotalForce[i]+=nodeForce
            for t in active2Triangles[i]:
                force2Triangles(i,t)
            for t in active1Triangles[i]:
                force1Triangles(i,t)

        ## DEBUG
        if not O.iter%(int(6000*iterAdjustFactor)):
            debugFunc()

        if O.iter < 2:
            print("GridNode Kn: " +str(O.bodies[nodeIds[0][len(nodeIds[0])/2]].intrs()[0].phys.kn))
            print("Mesh height: "+str(heightM/(2*nh-2)))+ "   Mesh width: "+str(widthM1/(nw-1))
            print("GridNode mass: "+str(O.bodies[nodeIds[0][len(nodeIds[0])/2]].state.mass))
            # activeNodeMassM1=0
            # for n in activeNodes[0]:
            #     activeNodeMassM1+= O.bodies[n].state.mass
            # print ("Total active node mass (not incl connections!): "+str(activeNodeMassM1))
            debugFunc()


    def numActiveNodes(i,t):
        out=[]
        num=0
        for n in t:
            if isInsideRigidWalls(n):
                activeNodes[n] = 1  # put in dict to mark as active
                num+=1
            else:
                out.append(n)
                #nodeActive[n]=1  # needed for calc1TriArea and calc2TriArea
        if num==1 or num==2:
            for n in out:
                if n not in nodeOutsideDist:
                    distNodeOutside(n)

        return num


    def distNodeOutside(n):
        global nodeOutsideDist
        pos = O.bodies[n].state.pos
        y = 0; z = 0
        # check against all walls
        yMin = wallYMin.state.pos[1] - insidePadding
        yMax = wallYMax.state.pos[1] + insidePadding
        zMin = bottomPlate.state.pos[2] - insidePadding
        zMax = topPlate.state.pos[2] + insidePadding
        if yMin > pos[1]: y = yMin-pos[1]
        if yMax < pos[1]: y = pos[1]-yMax
        if zMin > pos[2]: z = zMin-pos[2]
        if zMax < pos[2]: z = pos[2]-zMax
        nodeOutsideDist[n]=[y,z]


    def force1Triangles(i,t):
        inPos=0
        outPos=[]
        ins=0
        out=[]
        outDist=[]
        for n in t:
            if n in nodeOutsideDist:
                out.append(n)
                outPos.append(O.bodies[n].state.pos)
                outDist.append(nodeOutsideDist[n])
            else:
                inPos = O.bodies[n].state.pos
                ins = n
                #inDist = insideDist(n)

        # nonZeroDivision(a,b) returns a/b or 0 if b==0
        # calc % of full length from out node to in nodes for the inactive portion of the vector
        part1 = max(nonZeroDivision(outDist[0][0],abs(inPos[1]-outPos[0][1])), nonZeroDivision(outDist[0][1], abs(inPos[2]-outPos[0][2])))
        part2 = max(nonZeroDivision(outDist[1][0],abs(inPos[1]-outPos[1][1])), nonZeroDivision(outDist[1][1],abs(inPos[2]-outPos[1][2])))

        # Vectors from inside nodes to outside nodes
        outIn1 = inPos - outPos[0]; outIn2 = inPos - outPos[1]
        # Vectors pointing from inside to the walls -> active area
        inV1 = outIn1 * (1-part1) ; inV2 = outIn2 * (1-part2)
        if not pressureNormToMem:  # ignore the x-coord -> projection on the y-z-plane
            inV1[0]=inV1[0]=0

        crossVec = 0.5 * inV1.cross(inV2)

        if pressureNormToMem:
            # ensure correct direction of crossVec
            if i == 0:
                if crossVec[0] > 0:
                    crossVec *= -1
            elif i == 1:
                if crossVec[0] < 0:
                    crossVec *= -1
            for n in t:
                O.forces.addF(n,
                              sigmaX * crossVec * 1 / 3)  # 1/3 for even distr of force to nodes -> This should be different
            # add force of all nodes:
            activeTotalForce[i] += -crossVec.norm() * sigmaX
        else:
            nodeForce = -crossVec.norm() * sigmaX * 1 / 3
            for n in t:
                O.forces.addF(n, nodeForce * dirVector[i])  # 1/3 for even distr of force to nodes -> This should be different
            # add force of all nodes:
            activeTotalForce[i] += nodeForce * 3

    def force2Triangles(i,t):
        inPos = []  # inside nodes
        outPos = 0   # outside node
        ins = []
        out = 0
        for n in t:
            if n in nodeOutsideDist: # if n exist in dict
                out=n
                outPos=O.bodies[n].state.pos
                outDist = nodeOutsideDist[out]  # [y-dist to zWall, z-dist to yWall]
            else:
                ins.append(n)
                inPos.append(O.bodies[n].state.pos)
        # nonZeroDivision(a,b) returns a/b or 0 if b==0
        # calc % of full length from out node to in nodes for the inactive portion of the vector
        part1 = max(nonZeroDivision(outDist[0],abs(inPos[0][1]-outPos[1])), nonZeroDivision(outDist[1], abs(inPos[0][2]-outPos[2])))
        part2 = max(nonZeroDivision(outDist[0],abs(inPos[1][1]-outPos[1])), nonZeroDivision(outDist[1],abs(inPos[1][2]-outPos[2])))
        # Vectors from inside nodes to outside nodes
        outIn1 = outPos - inPos[0]; outIn2 = outPos - inPos[1]
        # Vectors pointing from outside to the walls -> use to remove inactive area
        negV1 = outIn1 * part1 ; negV2 = outIn2 * part2
        if not pressureNormToMem:  # ignore the x-coord -> projection on the y-z-plane
            outIn1[0]=outIn2[0]=negV1[0]=negV2[0]=0

        # 0.5 ( tot area triangle vector - area outside vector ) # Works because they are lin. dependent
        crossVec = 0.5 * ( outIn1.cross(outIn2) - negV1.cross(negV2) )

        if pressureNormToMem:
            # ensure correct direction of crossVec
            if i == 0:
                if crossVec[0] > 0:
                    crossVec *= -1
            elif i == 1:
                if crossVec[0] < 0:
                    crossVec *= -1
            for n in t:
                O.forces.addF(n, sigmaX * crossVec * 1 / 3 )  # 1/3 for even distr of force to nodes -> This should be different!
            # add force of all nodes:
            activeTotalForce[i] += -crossVec.norm()*sigmaX
        else:
            nodeForce = -crossVec.norm()*sigmaX*1/3
            for n in t:
                O.forces.addF(n, nodeForce*dirVector[i])  # 1/3 for even distr of force to nodes -> This should be different!
            # add force of all nodes:
            activeTotalForce[i] += nodeForce*3


    def nonZeroDivision(n,d):
        return n/d if d else 0

    def calc3TriArea(t):  # 3 active nodes
        pos1=O.bodies[t[0]].state.pos
        pos2=O.bodies[t[1]].state.pos
        pos3=O.bodies[t[2]].state.pos
        pos1[0]=pos2[0]=pos3[0]=0  # disregard x-coord
        v1=pos1-pos2; v2=pos1-pos3
        # calc area as 1/2 crossproduct
        return 0.5 * v1.cross(v2).norm()

    if pressureNormToMem:  # overwrite function above
        def calc3TriArea(i,t):
            pos1 = O.bodies[t[0]].state.pos
            pos2 = O.bodies[t[1]].state.pos
            pos3 = O.bodies[t[2]].state.pos
            v1 = pos1 - pos2; v2 = pos1 - pos3

            crossVec =  0.5 * v1.cross(v2)
            if i==0:
                if crossVec[0] > 0:
                    crossVec*=-1
            elif i==1:
                if crossVec[0] < 0:
                    crossVec*=-1
            # normal area vector
            return crossVec


    def activeAreaX():  # sample area in contact with membrane in x-dir + padded area
        return (wallYMax.state.pos[1]-wallYMin.state.pos[1]+2*insidePadding)*\
               (topPlate.state.pos[2]-bottomPlate.state.pos[2]+2*insidePadding)

    if activeNodesPadding:
        insidePadding = gridColW*1.2
    else:
        insidePadding=0
    def isInsideRigidWalls(nodeId):
        pos = O.bodies[nodeId].state.pos
        # check against all walls
        if wallYMin.state.pos[1] - insidePadding > pos[1]: return False
        if wallYMax.state.pos[1] + insidePadding < pos[1]: return False
        if bottomPlate.state.pos[2] - insidePadding > pos[2]: return False
        if topPlate.state.pos[2] + insidePadding < pos[2]: return False
        return True


    def debugFunc():
        print("Timestep: " + str(O.dt))
        print("Prescribed stress: " + str(-sigmaX) + "      Applied stress: " + str(activeTotalForce[0] / activeAreaX()))


# preallocate
reducedForceTop=[0,0,0,0]
reducedForceBot=[supportForce*2./3, supportForce*1./3]  # no change due to deformation

reduction = 0
# deactivate force on membrane as wall passes -> Pyrunner loop in engines
def updateReducedLoad():
    reduction = activeTopFacetRatio()  # from 1 to 0 dep on load plate height
    if borderNodesHalfLoad==0:  # whether to use 1/2 on border nodes or thirds
        reducedForceTop[0] = supportForce * 1./3 * (1 + 2 * reduction)    # 1 at max and 1/3 at min
        reducedForceTop[1] = supportForce * 2./3 * (1 + 0.5 * reduction)  # 1 at max and 2/3 at min
        # top active nodes
        reducedForceTop[2] = supportForce * 1./3 * reduction              # 1/3 at max and 0 at min
        reducedForceTop[3] = supportForce * 2./3 * reduction              # 2/3 at max and 0 at min
    else:
        reducedForceTop[0] = supportForce * 0.5 * (1 + reduction)    # 1 at max and 1/2 at min
        reducedForceTop[1] = supportForce * 0.5 * (1 + reduction)    # 1 at max and 1/2 at min
        # top active nodes
        reducedForceTop[2] = supportForce * 0.5 * reduction          # 1/2 at max and 0 at min
        reducedForceTop[3] = supportForce * 0.5 * reduction          # 1/2 at max and 0 at min


reductionTopRow=1
reductionSecRow=1
def updateReducedLoadMoreNW():
    global reductionTopRow
    global reductionSecRow
    reductionTopRow = activeTopFacetRatio()        # ranges from 1 to 0
    reductionSecRow = 0.5+0.5*activeTopFacetRatio()  # ranges from 1 to 1/2


# use second row of top nodes' position to determine the ratio of covered top active pfacet
def activeTopFacetRatio():
    factor=(topPlate.state.pos[2]-O.bodies[topNodes[0][0]].state.pos[2])/pFacetHeight
    if factor>1:
        return 1
    elif factor<0:
        print("activeTopFacetRatio returnerer negativ verdi!")
    else:
        return factor

# After initial tiny steps to sync forceEngine, set a reasonable timestep
ts.timestepSafetyCoefficient = adjustTimestepSafety

###########################################
#####   Strain control  #####
###########################################

def checkUnbalanced():
    if compress == 0: return
    # at the very start, unbalanced force can be low as there is only few contacts,
    # but it does not mean the packing is stable
    if rigidWalls==1:
        if O.iter < (iterTargetFlex) : return    # 15000+5*numParticles
    if rigidWalls == 2:
        if O.iter < (iterTargetRigid): return

    #if abs((yMaxWallStress()-sigmaY)/sigmaY) > 0.05: return

        # the rest will be run only if unbalanced is < .1 (stabilized packing)
    print ("Unbalanced force = ",unbalancedForce())
    if unbalancedForce()>1: return

    # reset reference positions before starting deviatoric compression
    for i in sphereIds:  # sphereIds includes clumps and clumpMembers if clumpSample
        s = O.bodies[i]
        s.state.refPos = s.state.pos
        s.state.refOri = s.state.ori

    global startedStrain
    startedStrain=1  # disables top and bot plate stress controls

    # reset height after stress control might have changed it to get real strain values
    global height
    resetStrain = verticalStrain()
    height = topPlate.state.pos[2]-bottomPlate.state.pos[2]
    print("starting compression with height="+str(height)+" -> "+str(resetStrain*100)+"% vert strain from original dim")

    addPlotData() # save one data point with stable state graph data before starting straining

    # constant loadplate velocity
    if strainBothWays:
        global wallVelocity
        wallVelocity*=0.5
        bottomPlate.state.vel=(0,0,wallVelocity)
    else:
        bottomPlate.state.vel=(0,0,0)
    topPlate.state.vel = (0, 0, -wallVelocity)

    if topCoverFollowStrain:
        for n in topCoverNodeIds:
            O.bodies[n].state.vel=Vector3(0,0,-wallVelocity)
        if strainBothWays:
            for n in botCoverNodeIds:
                O.bodies[n].state.vel = Vector3(0, 0, +wallVelocity)

    # Swap function to stop it from creating boxes and plates
    checker.command='strainTracker()'


verticalStrainTracked=[]
strainSelecter = 0
strainIncrements.append(1)  #100% strain -> dummy value to stop call of displacementTracker() after reaching last target
def strainTracker():
    if captureParaviewData and captureData:
        global strainSelecter  # make global
        # check strain intervals and save displacement data at specified strain
        if verticalStrain() > strainIncrements[strainSelecter]:
            print("Saving vtk-data at strain: ", verticalStrain())
            paraviewData()
            verticalStrainTracked.append(verticalStrain())  # for checking against specified strain
            ## proceed to next strain target
            if strainSelecter < len(strainIncrements) - 1:  # to not run out of bounds for the list strainIncrements
                strainSelecter += 1

    if verticalStrain() > -targetStrain:
        # topPlate.state.vel = (0, 0, 0)
        # bottomPlate.state.vel = (0, 0, 0)
        checker.dead = True
        finished()


# control velocities dep on volume change potential:
controlVelX = widthX/(widthX+widthY+height) * wallVelocity*30
controlVelY = widthY/(widthX+widthY+height) * wallVelocity*30
controlVelZ = height/(widthX+widthY+height) * wallVelocity*10
startedStrain = 0  # while False: stress control of top and bot is active

poissonScaleFactor = 0.75  # how much sideCovers are moved of rigid y walls movements
def rigidBoundaryController():
    if rigidWalls == 0:
        return
    # Adjust top and bot plate while stabilizing sample
    if not startedStrain:
        stress1=botWallStress()
        stress2=topWallStress()
        vel1 = (1 - abs(stress1 / sigmaZ)) * controlVelZ
        vel2 = -(1 - abs(stress2 / sigmaZ)) * controlVelZ
        bottomPlate.state.vel[2] = vel1
        topPlate.state.vel[2] = vel2
    if rigidWalls != 0 and not planeStrainConditions:
        stress1=yMinWallStress()
        stress2=yMaxWallStress()
        vel1 = (1 - abs(stress1 / sigmaY)) * controlVelY
        vel2 = -(1 - abs(stress2 / sigmaY)) * controlVelY
        wallYMin.state.vel[1] = vel1
        wallYMax.state.vel[1] = vel2
        if rigidWalls==1 and genSideCovFacets == 1:
            # sidecovers follow walls
            for n in sideCoverNodeIds[0][0]:
                O.bodies[n].state.vel[1] = vel1
                if moveCoversWithYWalls:  # approximate lateral strain also for covers with membrane
                    O.bodies[n].state.vel[0]= 0.5 * (vel1 - vel2) * poissonScaleFactor
            for n in sideCoverNodeIds[1][0]:
                O.bodies[n].state.vel[1] = vel1
                if moveCoversWithYWalls:
                    O.bodies[n].state.vel[0]= -0.5 * (vel1 - vel2) * poissonScaleFactor

            for n in sideCoverNodeIds[0][1]:
                O.bodies[n].state.vel[1] = vel2
                if moveCoversWithYWalls:
                    O.bodies[n].state.vel[0]= -0.5 * (-vel1 + vel2) * poissonScaleFactor
            for n in sideCoverNodeIds[1][1]:
                O.bodies[n].state.vel[1] = vel2
                if moveCoversWithYWalls:
                    O.bodies[n].state.vel[0] = 0.5 * (-vel1 + vel2) * poissonScaleFactor

    if rigidWalls == 2:
        stress1 = xMinWallStress()
        stress2 = xMaxWallStress()
        vel1 = (1 - abs(stress1 / sigmaX)) * controlVelX
        vel2 = -(1 - abs(stress2 / sigmaX)) * controlVelX
        wallXMin.state.vel[0] = vel1
        wallXMax.state.vel[0] = vel2

###############################
###### Support functions ######
###############################

# set velocity of membranes to zero
def membraneCalm():
    for i in range(numMembranes):
        for n in nodeIds[i]:
            O.bodies[n].state.vel = Vector3(0,0,0)
            ## try to stop sideways deformation
            # if i<2:
            #     O.bodies[n].state.vel[1]=0
            # elif i>=2:
            #     O.bodies[n].state.vel[0]=0

def elapsedTime():
    print('Elapsed time in minutes: ' + str((time.time() - start_time) / 60))


def nodeIdGetZ(nodeId):
    return O.bodies[nodeIds].state.pos[2]

## Functions for normal stress on confining facets
def topWallStress():
    wallForce = O.forces.f(topPlate.id)[2]  # resultant force in z-dir
    # wallShearX = O.forces.f(bottomPlate.id)[0]
    # wallShearY = O.forces.f(bottomPlate.id)[1]
    # engineering stress  F/A0
    if engiStress:
        return wallForce/(widthX*widthY)  # Using the undeformed geometry as loaded from sampleGen
    else:
        if rigidWalls==1:
            return wallForce / (widthX * widthY*(1-yRigidStrain()))
        if rigidWalls==2:
            return wallForce / (widthX*(1-xRigidStrain()) * widthY*(1-yRigidStrain()))

def botWallStress():
    wallForce = -O.forces.f(bottomPlate.id)[2]  # resultant force in z-dir
    # engineering stress  F/A0
    if engiStress:
        return wallForce / (widthX * widthY)  # Using the undeformed geometry as loaded from sampleGen
    else:
        if rigidWalls == 1:
            return wallForce / (widthX * widthY * (1 - yRigidStrain()))
        if rigidWalls == 2:
            return wallForce / (widthX * (1 - xRigidStrain()) * widthY * (1 - yRigidStrain()))

def yMinWallStress():
    wallForce = -O.forces.f(wallYMin.id)[1]  # resultant force in y-dir
    # engineering stress  F/A0
    if engiStress or rigidWalls==1:
        return wallForce/(widthX*height*(1-verticalStrain()))
    if rigidWalls==2:
        return wallForce / (widthX*(1 - xRigidStrain()) * height * (1 - verticalStrain()))

def yMaxWallStress():
    wallForce = O.forces.f(wallYMax.id)[1]  # resultant force in y-dir
    # engineering stress  F/A0
    if engiStress or rigidWalls==1:
        return wallForce/(widthX*height*(1-verticalStrain()))  # accounting for vert strain
    if rigidWalls==2:
        return wallForce / (widthX * (1 - xRigidStrain()) * height * (1 - verticalStrain()))

def xMinWallStress():
    wallForce = -O.forces.f(wallXMin.id)[0]  # resultant force in x-dir
    if engiStress:
        # engineering stress  F/A0
        return wallForce / (widthY * height * (1 - verticalStrain()))
    else:
        # accounting for vert and y strain
        return wallForce / (widthY*(1-yRigidStrain()) * height*(1-verticalStrain()))

def xMaxWallStress():
    wallForce = O.forces.f(wallXMax.id)[0]  # resultant force in x-dir
    if engiStress:
        # engineering stress  F/A0
        return wallForce / (widthY * height * (1 - verticalStrain()))
    else:
        # accounting for vert and y strain
        return wallForce/(widthY * (1 - yRigidStrain()) * height * (1 - verticalStrain()))


def verticalStrain():
    return 1-(topPlate.state.pos[2]-bottomPlate.state.pos[2])/height

def yRigidStrain():  # positive strain in compression
        return 1 - (wallYMax.state.pos[1]-wallYMin.state.pos[1])/widthY

def xRigidStrain():  # positive strain in compression
    if rigidWalls==1:  # no attempt at estimating membrane strain
        return 0
    elif rigidWalls==2:
        return 1 - (wallXMax.state.pos[0] - wallXMin.state.pos[0]) / widthX


def samplePorosity(res):
    #res = 1000  # number of voxels used to check designated area for overlap with particles to calc porosity
    safetyMargin=0.1*widthX  # pull volume this much back from each membrane border to ensure homogeneity
    startPoint = Vector3(wallStartPos[0],wallYMin.state.pos[1],bottomPlate.state.pos[2]) + Vector3(safetyMargin,0,0)
    endPoint = Vector3(wallStartPos[1],wallYMax.state.pos[1],topPlate.state.pos[2]) -Vector3(safetyMargin,0,0)

    # end and start define the corners of the cubic volume to check porosity of
    return voxelPorosity(resolution = res, start = startPoint, end = endPoint )
    # "The calculation method is to divide whole volume into a dense grid of voxels (at given resolution),
    # and count the voxels that fall inside any of the spheres. This method allows one to calculate porosity in any given sub-volume of a whole sample"
    #->Use this to calc a representative porosity in an internal volume of the sample, circumventing the problem of flexible borders?


def accumulatedRotation(bodyId):
    global accuRotationsDict
    if 'accuRotationsDict' not in globals():
        accuRotationsDict={}
    if accuRotationsDict.has_key(bodyId):
        accuRotationsDict[bodyId]+= O.bodies[bodyId].state.rot().norm()
    else:
        accuRotationsDict[bodyId] = O.bodies[bodyId].state.rot().norm()
    return accuRotationsDict[bodyId]
# define acculatedRotation() as builtin to allow access while in the vtkExporter scope
__builtin__.accumulatedRotation=accumulatedRotation


def deltaContactNum(bodyId):  # find change in contact number after stable sample
    global contactNumInit
    if 'contactNumInit' not in globals():  # initialize dict in global scope
        contactNumInit = {}
    if contactNumInit.has_key(bodyId):
        return len(O.bodies[bodyId].intrs())-contactNumInit[bodyId]
    else:
        contactNumInit[bodyId]=len(O.bodies[bodyId].intrs())  # set initial contact number to measure change later
        return 0
# make accessible in the vtkExporter scope
__builtin__.deltaContactNum = deltaContactNum


def deltaContactNumEachStep(bodyId):
    global lastContactNum
    if 'lastContactNum' not in globals():
        lastContactNum = {}
    if lastContactNum.has_key(bodyId):
        deltaNumC = len(O.bodies[bodyId].intrs())-lastContactNum[bodyId]
        lastContactNum[bodyId] = len(O.bodies[bodyId].intrs()) # set current contact num for next iteration
        return deltaNumC
    else:
        lastContactNum[bodyId]=len(O.bodies[bodyId].intrs())
        return 0
# make accessible in the vtkExporter scope
__builtin__.deltaContactNumEachStep = deltaContactNumEachStep



if tesselatedMicroStrain:
    # Initialize
    tesselationStep = 0
    tesselationCounter = 1
    TW = TesselationWrapper()
def tesselatedStrain():
    global tesselationStep
    global tesselationCounter
    if tesselationStep == 0:
        TW.triangulate()  # compute regular Delaunay triangulation, don’t construct tesselation
        print
        "triangulated"
        TW.computeVolumes()  # will silently tesselate the packing, then compute volume of each Voronoi cell
        print
        "volumes computed"
        TW.setState(0)
        print
        "state 0 set"
        # prepare for next step
        tesselationStep = 1

    elif tesselationStep == 1:
        TW.setState(1)
        TW.defToVtk(path + "strain.vtk")
        print
        "saved tesselated strain!"
        # prepare for next step
        tesselationStep = 0


vtkFileNumber = 0
localRes = 50  # resolution for localised volume -> 50 is min value

hRes = 40                               # pixels in z-dir
wRes = int(round(widthX/height*hRes))   # pixels in x-dir
poroThick = 0.2 * widthY                # y-dim of checked center strip
defaultPorosityList = []                # use to find changes in porosity
firstTime = 1
def localisedPorosity():
    global firstTime, defaultPorosityList
    # Check porosity in a grid search through the center of the sample - in the x-z plane due to sigmaX<sigmaY
    poroAndPosList = [[],[]]  # reset [position, porosity] lists
    currentHeight = topPlate.state.pos[2]-bottomPlate.state.pos[2]
    heightB = currentHeight/hRes  # height of checked "pixel"-box
    widthB = widthX/wRes  # width of checked box
    #wRes = int(round(widthX/widthB)) #-> problem: changes the number of boxes across the width! Use predetermined external wRes instead
    midY = (wallYMin.state.pos[1]+wallYMax.state.pos[1])/2
    startRef = Vector3(wallStartPos[0], midY-poroThick/2, bottomPlate.state.pos[2])
    endRef = startRef + Vector3(widthB, poroThick, heightB)

    for h in range(hRes):
        for w in range(wRes):
            start = startRef + Vector3(w * widthB, 0, h * heightB)
            end = endRef + Vector3(w * widthB, 0, h * heightB)
            centerSquare = start + Vector3(widthB/2,poroThick/2, heightB/2)  # halfway to end pos

            poroAndPosList[0].append([centerSquare[0],centerSquare[1],centerSquare[2]])
            poroAndPosList[1].append(calcPorosity(localRes, start, end))

    if firstTime:
        defaultPorosityList=poroAndPosList[1]
        firstTime=0
    else:
        poroAndPosList[1] = [por1-por0 for por1,por0 in zip(poroAndPosList[1],defaultPorosityList)]
            #poroAndPosList[1]-defaultPorosityList
        porosityVTKCreator(poroAndPosList,heightB)

def calcPorosity(res,start,end):
    return voxelPorosity(resolution=res, start=start, end=end)



def porosityVTKCreator(poroList,heightB):
    global vtkFileNumber
    points = poroList[0]
    porosity = poroList[1]
    heightB = [heightB]*len(points)
    structure = pyvtk.PolyData(points=points)
    pointdata = pyvtk.PointData(
        pyvtk.Scalars(heightB, name="heightB", lookup_table="default"),
        pyvtk.Scalars(porosity, name="porosity", lookup_table="default"),
    )

    vtk = pyvtk.VtkData(structure, pointdata, "Localised porosity")
    vtkFileName = path+"localPorosity"+str(vtkFileNumber)
    vtk.tofile(vtkFileName, 'ascii')
    vtkFileNumber+=1


##############################
#####   Plotting/data  #####
##############################

def printParam():

    print ("numParticles =",numParticles,"\tYoung =",young/1e6,"* e6")
    print ("sigmaIso =", sigmaIso)
    print ("wallVel =",wallVelocity)
    print ("calmIterPeriod =",calmIterPeriod)

    print ("PfacetDensity =",densityFacet,"\tGridNodeDensity = ",densityGridNode)
    print ("membraneStiff =",membraneStiff)

    infoPrinter.command='printProgress()'

extremesNorm= (aabbExtrema()[0]-aabbExtrema()[1]).norm()  # initial extremes
def printProgress():
    # check if membrane slips away:
    lower, upper = aabbExtrema()
    if (upper - lower).norm() > extremesNorm * 2:
        print ("Membrane unstable -> failure")
        print ("Vert strain ", verticalStrain() * 100, "%")
        elapsedTime()
    else:
        print ("Vert strain ",verticalStrain()*100,"%")
        elapsedTime()
        sys.stdout.flush()  # flush print to output (for batch)


def addPlotData():
    if not startedStrain:
        plot.addData(unbalanced=unbalancedForce(), i=O.iter,)
        return  # do not add any more data
    else:
        if rigidWalls==2:
            plot.addData(unbalanced=unbalancedForce(), i=O.iter,
                         ezz=verticalStrain(), szzTop=topWallStress(),
                         szzBot=botWallStress(), syyMin=yMinWallStress(),
                         syyMax=yMaxWallStress(),sxxMin=xMinWallStress(), sxxMax=xMaxWallStress(),
                         n=samplePorosity(poroResolution),
                         )
        else:
            plot.addData(unbalanced=unbalancedForce(), i=O.iter,
                         ezz=verticalStrain(), szzTop=topWallStress(),
                         szzBot=botWallStress(), syyMin=yMinWallStress(),
                         syyMax=yMaxWallStress(), sxxMin = -sigmaX, sxxMax= -sigmaX,   # set sxxMin and sxxMax as the prescribed stress for stress path plotting
                         n=samplePorosity(poroResolution),
                         # syy =    ,# HORISONTAL STRESS on walls
                         # sxx = sigmaIso, # ? Difficult to approx. Maybe sum y force of all nodes/Area -> should be close to zero since applied force is balanced by sample interaction?
                         # qCrit = szz - sxx  # crit because sxx<syy ? -> Thus shearband orientation in x-z plane? -> Out of membrane
                         # # ev=-triax.strain[0]-triax.strain[1]-triax.strain[2],
                         # sxx=-triax.stress(triax.wall_right_id)[0],
                         # syy=-triax.stress(triax.wall_top_id)[1],
                         # szz=-triax.stress(triax.wall_front_id)[2],
                         # # save all available energy data
                         # Etot=O.energy.total(), n=triax.porosity, **O.energy
                         )


if plotData:  # create plots
    if rigidWalls==1:
        plot.plots={

        'ezz':('szzTop','szzBot','syyMax','syyMin','sxxMax','sxxMin'),
        "i":('unbalanced'),
        'ezz ': ('n'),  # 'ezz ' -> use a space in the string to stop from overwriting 'ezz':stress data
        }
    elif rigidWalls==2:
        plot.plots = {
            'ezz':('szzTop','szzBot','syyMax','syyMin','sxxMax','sxxMin'),
            'i':('unbalanced'),
            'ezz ':('n'),
                      }
if showPlot:
    plot.plot()  # show the created plots

##############################
#####   Run simulation  #####
##############################


# if not using GlobalStiffnessTimeStepper()
#O.dt = PWaveTimeStep()*timestepSafety*0.008

# enable 3D view
if view3D:
    qt.View()


# Save sample after setup?
O.saveTmp()


#########################
#### CALM FUNCTION ####
########################
# To stabilize sample when using replaceByClumps -> swapping in clumps can lead to overlaps between spheres
if calmFunction==1:
    print('\nActivated continuous calm() function\n')
    ## calm all bodies
	# O.engines=O.engines+[PyRunner(iterPeriod=calmIterPeriod,command='calm()',label='calmRunner')]
    ## just the membrane
    O.engines=O.engines+[PyRunner(iterPeriod=calmIterPeriod,command='membraneCalm()',label='calmRunner')]
	# if O.iter>
	#     calmRunner.dead=True

elif calmFunction==2:
    print('\nPlease wait a minute ...\n')
    ## calm all bodies
    O.engines = O.engines + [PyRunner(iterPeriod=10, command='calm()', label='calmRunner')]
    ## just the membrane
    #O.engines = O.engines + [PyRunner(iterPeriod=10, command='membraneCalm()', label='calmRunner')]

    O.run(calmSteps, 1)  # default 1000000   evt skalere med young??
    O.pause()
    calmRunner.dead = True


# run/stop condition
if not view3D:
    O.run()


#######################################
#### Saving Simulation Parameters ####
#######################################
if captureData:
    # copy contents of paramFile from sampleGen
    paramFile=open("1paramSampleGen.txt","r")
    logFileName = "1SimulationData.txt"
    dataF=open(path+logFileName,"w")
    dataF.write(paramFile.read())
    paramFile.close()

    dataF.write("\n\nCOMPRESSION SIMULATION WITH MEMBRANE:")
    dataF.write("\nfinalFricDegree=" + str(finalFricDegree))
    dataF.write("\nwallVelocity="+str(wallVelocity)+"\nmembraneStiff="+str(membraneStiff)+"\nnw="+str(nw)+"; nh ="+str(nh))
    dataF.write("\nyoungFacet="+str(youngFacet)+"\ndensityFacet="+str(densityFacet)+"\ndensityGridNode="+str(densityGridNode))
    dataF.write("\ndamping="+str(damping)+"\ndensityScaleFactor="+str(densityScaleFactor))
    dataF.write("\nadjustTimestepSafety=" + str(adjustTimestepSafety))


    dataF.write("\n\nSETTINGS\nlockAnchorsZ="+str(lockAnchorsZ)+"\nrigidWalls="+str(rigidWalls))
    dataF.write("\nstrainBothWays=" + str(strainBothWays))
    dataF.write("\npressureNormToMem=" + str(pressureNormToMem) + "\nactiveNodesPadding="+str(activeNodesPadding))
    dataF.write("\nmoveCoversWithWalls=" + str(moveCoversWithYWalls))

    dataF.close()  # close file to save data

def finished():
    print ('Reached ', verticalStrain() * 100, '% strain')
    elapsedTime()
    print("Iterations: ",O.iter)

    # Stop vertical straining
    topPlate.state.vel = (0, 0, 0)
    bottomPlate.state.vel = (0, 0, 0)

    # save plotData
    if plotData and captureData and not batchMode:
        plot.saveDataTxt(path + "1results" + date + clock + ".txt")
        # create plotted graph files
        plot.saveGnuplot(path+O.tags['id'][4:])
        plottingFileName = O.tags['id'][4:] + ".data.bz2"
        f = open(path+"1dataNames.txt", "a+")
        # add identifying data for plotting
        f.write(str(plottingFileName) + '\t' + str(int(compFricDegree)) + '\t' + str(finalFricDegree) + '\t' + str(
            young) + '\t' + str(wallVelocity) + '\t' + str(densityScaleFactor) + '\t' + str(damping) + '\n')
        # compactedPorosity, densityScaling, young,
        f.close()

    # save final simulation time
    if captureData:
        dataF = open(path + logFileName, "a+")  # a+ -> open and append to file/create and write to file

        dataF.write("\n\nSimulation time = " + str((time.time() - start_time) / 60) + " minutes")

        dataF.close()

    O.pause()  # last line due to batch terminating when pausing simulation?

###################################
######## TESTING FUNCTIONS ########
###################################
# Manual data saving for sample instability
def saveData():
    plot.saveGnuplot(path + O.tags['id'][4:])
    plottingFileName = O.tags['id'][4:] + ".data.bz2"
    f = open(path + "1dataNames.txt", "a+")
    # add identifying data for plotting
    f.write(str(plottingFileName) + '\t' + str(int(compFricDegree)) + '\t' + str(finalFricDegree) + '\t' + str(
        young) + '\t' + str(wallVelocity) + '\t' + str(densityScaleFactor) + '\t' + str(damping) + '\n')
    # compactedPorosity, densityScaling, young,
    f.close()
    print("Saved plot data")

# send membranes away from sample
def removeMembranes():
    confiningEngine.dead=True
    for i in range(numMembranes):
        for n in nodeIds[i]:
            O.bodies[n].state.vel[0]=-100*pow(-1,i)


## Testing function for overriding compress=0
def compressIt():
    # constant loadplate velocity
    if strainBothWays:
        global wallVelocity
        wallVelocity*=0.5
        bottomPlate.state.vel=(0,0,wallVelocity)
    topPlate.state.vel = (0, 0, -wallVelocity)

    if topCoverFollowStrain:
        for n in topCoverNodeIds:
            O.bodies[n].state.vel=Vector3(0,0,-wallVelocity)
        if strainBothWays:
            for n in botCoverNodeIds:
                O.bodies[n].state.vel = Vector3(0, 0, +wallVelocity)


    # Swap function to stop it from creating boxes and plates
    checker.command='strainTracker()'


def stressOnPlaneX():  # adds all forces in interactions that cross the center plane and divides by the intersected area
    # put the plane inside of membrane AND coverFacets to stop interference
    return forcesOnPlane(Vector3(wallStartPos[0]+covRad*2.5, 0.5*widthY, 0.5*height),Vector3(1,0,0))/(height*widthY)


#############



# For batch mode
utils.waitIfBatch()

if batchMode:
    plottingFileName = O.tags['id'][4:] + ".data.bz2"
    f = open("1dataNames.txt", "a+")
    # add identifying data for plotting
    f.write(str(plottingFileName)+'\t' + str(int(compFricDegree)) + '\t' + str(finalFricDegree) + '\t' + str(young) +'\t' + str(wallVelocity) +'\t' + str(densityScaleFactor) +'\t' + str(damping) + '\n')
    # compactedPorosity, densityScaling, young,
    f.close()
    print('gnuplot ' + str(plot.saveGnuplot(O.tags['id'][4:])))
    print("\nTag: "+str(O.tags["id"][-5:]))
